/*-
 * Copyright (C)2014 @BABOLO http://www.babolo.ru/
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static const char copyright[] = "\
@(#)Copyright (C)2014 @BABOLO http://www.babolo.ru/\n\
@(#)All rights reserved.\n\
@(#)V.M "VERS;
static const char rcsid[] = "$Id: auwithd.c,v 1.2 2014/09/16 17:27:45 babolo Exp $";
#endif /* not lint */

#define BLIN_COMPAT  3
#define Bpars_COMPAT 3

#include <machine/param.h>
#include <sys/types.h>
#include <sys/event.h>
#include <sys/time.h>
#include <sysexits.h>
#include <strings.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <stdio.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <babolo/BLINflag.h>
#include <babolo/parser.h>
#include "auwith_int.h"

static auwith_plug *auwith_allplug[] =
{ &auwith_def_base            /* !!! [0] only !!! */
, &auwith_def_sine
, &auwith_def_dspo
, &auwith_def_dspi
, &auwith_def_udpo
, &auwith_def_udpi
, &auwith_def_file
, NULL
};

static int
/*****************************************************************************************************/
plugins(auwith_opd *opd) {                                                                        /****
 *****************************************************************************************************/
    int ex = EX_OK;

    ifBLIN_QX3("+");
    opd->plug = auwith_allplug;
    for (opd->nextplugin = 0; !!opd->plug[opd->nextplugin]; ++opd->nextplugin);
# if 0
    ifBLIN_QX3("+");
    if  (!(opd->plug = malloc(sizeof(auwith_plug*) * opd->pluginsz))) {
        ifBLIN_QX0("Malloc failed");
        ERROUT(EX_OSERR, ENOMEM);
    }
    opd->plug[opd->nextplugin++] = &auwith_def_base;
    opd->plug[opd->nextplugin++] = &auwith_def_sine;
    opd->plug[opd->nextplugin++] = &auwith_def_dspo;
    opd->plug[opd->nextplugin++] = &auwith_def_dspi;
    opd->plug[opd->nextplugin++] = &auwith_def_udpo;
    opd->plug[opd->nextplugin++] = &auwith_def_udpi;
    opd->plug[opd->nextplugin++] = &auwith_def_file;
# endif
    ifBLIN_QX3("- %d", ex);
    return(ex);
}

static int
/*****************************************************************************************************/
parsin(auwith_opd *opd) {                                                                        /****
 *****************************************************************************************************/
    char *flar = "d::l:::qv=:?";
    int c, i, o, ex = EX_OK;

    ifBLIN_QX3("+");
    if  (!opd) {
        ifBLIN_QX0("no opd");
        ERROUT(EX_DATAERR, EFAULT);
    }
    if  (!opd->bos) {
        ifBLIN_QX0("no bos");
        ERROUT(EX_DATAERR, EFAULT);
    }
    if  (!!babolo_chopts(opd->bos, 0, flar)) {
        ifBLIN_QW0("chopts");
        EXOUT(EX_SOFTWARE);
    }
    ex = EX_OK;
    if  (opd->flags & AUWITH_DUMPOP) babolo_dumpopts(opd->bos);
    while (!!(c = babolo_getopts(opd->bos))) {
        switch (c) {
        case 'd': if  (!opd->pidm) {
                      ifBLIN_QX0("Use -f option before -d");
                      ERROUT(EX_USAGE, EPROTO);
                  }
                ; i = babolo_getoptunum(opd->bos, 0);
                ; o = babolo_getoptunum(opd->bos, 0);
                ; auwith_cella(opd->pidm, i, o)->flags &= ~AUWITH_USED;
                ; break;
        case 'l': if  (!opd->pidm) {
                      ifBLIN_QX0("Use -f option before -l");
                      ERROUT(EX_USAGE, EPROTO);
                  }
                ; i = babolo_getoptunum(opd->bos, 0);
                ; o = babolo_getoptunum(opd->bos, 0);
                ; auwith_cella(opd->pidm, i, o)->flags =
                      (babolo_getoptunum(opd->bos, 0) & AUWITH_LEVEL) | AUWITH_USED;
                ; break;
        case 'q': BLIN_QUIET(opd->flags);
                ; break;
        case 'v': BLIN_VERBOSE(opd->flags);
                ; break;
        case '=': switch(babolo_getoptsarg(opd->bos)[0]) {
                  case '0': blin_ctl(BLIN_CTL_DUMP);           exit(EX_OK);
                  case '1': babolo_optest(flar);               exit(EX_OK);
                  case '2': babolo_optext(opd->bos);           exit(EX_OK);
                  case '3': babolo_dumpopts(opd->bos);         exit(EX_OK);
                  }
                ; break;
        default : ifBLIN_QX0("Illegal flag -%c", opd->bos->c);
        case '?':
            fprintf( stderr
                   , "auwithd base V%u @BABOLO " DATE "\n"
                     "Usage: auwithd [-options] base [-options]\n"
                     "   -d in out       deconfig matrix cell [in, out]\n"
                     "   -E -X XX        use XX as env name for opt's arg\n"
                     "   -l in out level config matrix cell [in, out] with level\n"
                     "   -q              silent\n"
                     "   -v[v[v...]]     verbose\n"
                     "   -z TEXT         parce TEXT as arguments continuation\n"
                     "   -Z TEXT         parce TEXT as arguments continuation with env substitution\n"
                     "   -= mode         internal diagnostic\n"
                     "   -?              this text\n"
                   , auwith_def_base.mark.ver
                   )
            ;
    }   }
    ex = EX_OK;
out:
    ifBLIN_QX3("- %d", ex);
    return(ex);
}

_Noreturn static void
/*****************************************************************************************************/
usage(int ex) {                                                                                  /****
 *****************************************************************************************************/
    fprintf( stderr
           , "auwithd " VERS " @BABOLO " DATE "\n"
             "Usage: auwithd [-options]\n"
             "   -D              hard timing debug\n"
             "   -E -X XX        use XX as env name for opt's arg\n"
             "   -f file         control file\n"
             "   -i num          max inputs quantity\n"
             "   -L              local phase count\n"
             "   -M              look at matrix state\n"
             "   -o num          max outputs quantity\n"
             "   -O              dump opts structure\n"
             "   -P              max plugins quantity\n"
             "   -q              silent\n"
             "   -t time         scan matrix interval\n"
             "   -T file         unknown\n"
             "   -U              show config\n"
             "   -v[v[v...]]     verbose\n"
             "   -V              push verbosity to opts structure\n"
             "   -z TEXT         parce TEXT as arguments continuation\n"
             "   -Z TEXT         parce TEXT as arguments continuation with env substitution\n"
             "   -= mode         internal diagnostic\n"
             "   -?              this text\n"
           )
    ;
    exit(ex);
}

static int
/*****************************************************************************************************/
parsins(auwith_opd *opd, int argc, char *argv[]) {                                                /****
 *****************************************************************************************************/
    char *flar = "Df:i:LMo:OP:qt:T:UvV=:?";
    struct timespec tn;
    int c, ex = EX_OK;

    ifBLIN_QX3("+");
    if  (!(opd->bos = babolo_openopts(opd->flags | Bpars_SYME, 'Z'))) {
        ifBLIN_QW0("openopts");
        EXOUT(EX_SOFTWARE);
    }
    if  (!!babolo_setopts(opd->bos, opd->flags | Bpars_NONU, argc, argv, flar)) {
        ifBLIN_QW0("setopts");
        EXOUT(EX_SOFTWARE);
    }
    ex = EX_USAGE;
    if  (opd->flags & AUWITH_DUMPOP) babolo_dumpopts(opd->bos);
    while (!!(c = babolo_getopts(opd->bos))) {
        switch (c) {
        case 'D': opd->flags |= AUWITH_HARDTI;
                ; break;
        case 'f': if  (!(opd->pidf = babolo_getoptsarg(opd->bos))) goto out;
                ; break;
        case 'i': opd->inpachsz = babolo_getoptunum(opd->bos, 0);
                ; break;
        case 'L': opd->flags ^= AUWITH_CFAF;
                ; break;
        case 'M': opd->flags |= AUWITH_LOOK;
                ; break;
        case 'o': opd->outachsz = babolo_getoptunum(opd->bos, 0);
                ; break;
        case 'O': opd->flags |= AUWITH_DUMPOP;
                ; break;
        case 'P': opd->pluginsz = babolo_getoptunum(opd->bos, 0);
                ; break;
        case 'q': BLIN_QUIET(opd->flags);
                ; break;
        case 't': if  (!!(ex = auwith_totime(babolo_getoptsarg(opd->bos), &tn))) goto out;
                ; opd->tr = (u_int64_t)tn.tv_sec * 1000000000ULL + tn.tv_nsec;
                ; break;
        case 'T': if  (!(opd->T = babolo_getoptsarg(opd->bos))) goto out;
                ; break;
        case 'U': opd->flags |= AUWITH_DECONF;
                ; break;
        case 'v': BLIN_VERBOSE(opd->flags);
                ; break;
        case 'V': opd->bos->flags &= ~BLIN_MASK; opd->bos->flags |= opd->flags & BLIN_MASK;
                ; break;
        case '=': switch(babolo_getoptsarg(opd->bos)[0]) {
                  case '0': blin_ctl(BLIN_CTL_DUMP);           exit(EX_OK);
                  case '1': babolo_optest(flar);               exit(EX_OK);
                  case '2': babolo_optext(opd->bos);           exit(EX_OK);
                  case '3': babolo_dumpopts(opd->bos);         exit(EX_OK);
                  }
                ; break;
        case '?': usage(EX_OK);
        default : ifBLIN_QX0("Illegal flag -%c", opd->bos->c);
                ; usage(EX_USAGE);
    }   }
    ex = EX_OK;
out:
    ifBLIN_QX3("- %d", ex);
    return(ex);
}

static void
/*****************************************************************************************************/
deconf(auwith_opd *opd, auwith_io *io, FILE *fo) {                                               /****
 *****************************************************************************************************/
    u_int i, j;

#   define blin_internal_flags (opd->flags & BLIN_MASK)
    fprintf( fo, "auwithd %08X %s(%d),  %u plugged of %u, IN %u x %u OUT t=%u.%09u\n"
           , opd->flags
           , !!opd->pidf ? opd->pidf : ""
           , opd->pidd
           , opd->nextplugin
           , opd->pluginsz
           , opd->inpachsz
           , opd->outachsz
           , (u_int)(opd->tr / 1000000000ULL)
           , (u_int)(opd->tr % 1000000000ULL)
           );
    for (i = 0; i < opd->nextplugin; ++i) {
        for (j = 0; j < AUWITH_NAMESZ; ++j) {
            if  (!opd->plug[i]->pname[j]) fprintf(fo, " ");
              else fprintf(fo, "%c", opd->plug[i]->pname[j]);
        }
        fprintf( fo, " V%u %c%c%c%c%c%c"
               , opd->plug[i]->mark.ver
               , opd->plug[i]->mark.c0
               , opd->plug[i]->mark.c1
               , opd->plug[i]->mark.c2
               , opd->plug[i]->mark.c3
               , opd->plug[i]->mark.c4
               , opd->plug[i]->mark.c5
               );
        fprintf(fo, " ");
        auwith_capand(opd->plug[i]->flags, fo, 1);
        fprintf(fo, " ");
        auwith_caproc(opd->plug[i], fo);
        fprintf(fo, "\n");
    }
    fprintf(fo, " INPUTS  ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::\n");
    if  (!!opd->inpach) {
        for (i = 0; i < opd->inpachsz; ++i) {
            if  (!!opd->inpach[i]) {
                opd->inpach[i]->ent->plug->deconf(opd, opd->inpach[i], fo);
    }   }   }
    fprintf(fo, " OUTPUTS ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::\n");
    if  (!!opd->outach) {
        for (i = 0; i < opd->outachsz; ++i) {
            if  (!!opd->outach[i]) {
                opd->outach[i]->ent->plug->deconf(opd, opd->outach[i], fo);
    }   }   }
#   undef blin_internal_flags
}

static int
/*****************************************************************************************************/
activ(auwith_opd *opd, auwith_ent *ent, int kind) {                                              /****
 *****************************************************************************************************/
    auwith_bchain *bchain;
    auwith_io *si, *so;
    auwith_cell *cell;
    int ex = EX_OK;
    u_int i, o, k;

    ifBLIN_QX3("+ %d", kind);
    switch (kind) {
    case AUWITH_KIND_TIME:
        for (i = 0; i < opd->pidm->ninp; ++i) {
            for (o = 0; o < opd->pidm->nout; ++o) {
                cell = auwith_cella(opd->pidm, i, o);
                si = opd->inpach[i];
                so = opd->outach[o];
                if  (!si || !so) {
                    cell->flags = 0;
                    continue;
                }
                if  (!!(cell->flags & AUWITH_USED) && !!(cell->flags & AUWITH_ALGO)) {
                    for (k = 0, bchain = auwith_cadd(opd, cell)->bchain; bchain; bchain = bchain->next) ++k;
                    if  (k != cell->f.chainlen) {
                        ifBLIN_QX0("[%u %u] r%u v%u", i, o, k, cell->f.chainlen);
                        if  (65536 > k) cell->f.chainlen = k;
                    }
                    if  (cell->flags & AUWITH_CFAF) {
                        cell->f.faser = auwith_sfafer(&cell->f);
                        cell->f.doo = 1;
                    } else {
                        bcopy(&cell->f, &cell->t, sizeof(auwith_sfaf));
                        cell->t.doo = 1;
                }   }
                if  (!(cell->flags & AUWITH_USED) && !!(cell->flags & AUWITH_ALGO)) {
                    if  (!!si->ent->plug->passiv) {
                        if  (!!(ex = si->ent->plug->passiv(opd, cell, AUWITH_KIND_MIXDI))) {
                            ifBLIN_QW0("in passiv[%u] disable", i);
                            goto out;
                    }   }
                    if  (!!so->ent->plug->passiv) {
                        if  (!!(ex = so->ent->plug->passiv(opd, cell, AUWITH_KIND_MIXDI))) {
                            ifBLIN_QW0("out passiv[%u] disable", o);
                            goto out;
                    }   }
                    ex = auwith_handlers[(cell->flags & AUWITH_ALGO) >> AUWITH_ALGOS].fini(opd, cell);
                    if  (!!ex) {
                        ifBLIN_QW0("fini %u->%u", i, o);
                        goto out;
                    }
                    cell->flags &= ~AUWITH_ALGO;
                }
                if  (!!(cell->flags & AUWITH_USED) && !(cell->flags & AUWITH_ALGO)) {
                    cell->flags |= auwith_handle(opd, i, o);
                    /* XXXX flags from auwith_io ???? */
                    ex = auwith_handlers[(cell->flags & AUWITH_ALGO) >> AUWITH_ALGOS].init(opd, cell);
                    if  (!!ex) {
                        ifBLIN_QW0("init %u->%u", i, o);
                        goto out;
                    }
                    if  (!!si->ent->plug->passiv) {
                        if  (!!(ex = si->ent->plug->passiv(opd, cell, AUWITH_KIND_MIXEN))) {
                            ifBLIN_QW0("in passiv[%u] enable", i);
                            goto out;
                    }   }
                    if  (!!so->ent->plug->passiv) {
                        if  (!!(ex = so->ent->plug->passiv(opd, cell, AUWITH_KIND_MIXEN))) {
                            ifBLIN_QW0("out passiv[%u] enable", o);
                            goto out;
        }   }   }   }   }
        if  (!!(ex = auwith_queue(opd, NULL, ent, auwith_future(opd->tr)))) {
            ifBLIN_QW0("queue");
            goto out;
        }
        break;
    case AUWITH_KIND_KQ:
    default:
        ifBLIN_QX0("kind %d", kind);
        ERROUT(EX_USAGE, EINVAL);
    }
out:
    ifBLIN_QX3("- %d", ex);
    return(ex);
}

static int
/*****************************************************************************************************/
createcontrolfile(auwith_opd *opd) {                                                             /****
 *****************************************************************************************************/
    auwith_cell *cell;
    int ex = EX_OK;
    u_int i, o, j;
    ssize_t w, u;

    ifBLIN_QX3("+");
    u = sizeof(auwith_cell) * opd->inpachsz * opd->outachsz + sizeof(auwith_pid);
    u += sizeof(auwith_vol) * (opd->inpachsz + opd->outachsz);
    if  (0 > (opd->pidd = open(opd->pidf, O_RDWR | O_CREAT | O_TRUNC | O_EXLOCK, 0600))) {
        ifBLIN_QW0("open pid");
        EXOUT(EX_IOERR);
    }
    if  (!(opd->pidm = calloc(1, PAGE_SIZE))) {
        ifBLIN_QX0("Malloc failed #4");
        ERROUT(EX_OSERR, ENOMEM);
    }
    for (j = 0; u > 0; u -= PAGE_SIZE) {
        ++j;
        if  (0 > (w = write(opd->pidd, opd->pidm, PAGE_SIZE))) {
            ifBLIN_QW0("write");
            EXOUT(EX_IOERR);
        } else if (w != PAGE_SIZE) {
            ifBLIN_QX0("write %u <> %u", w, PAGE_SIZE);
            ERROUT(EX_IOERR, EIO);
    }   }
    free(opd->pidm);
    j *= PAGE_SIZE;
    opd->pidm = mmap(NULL, j, PROT_READ | PROT_WRITE, MAP_NOSYNC | MAP_SHARED, opd->pidd, 0);
    if  (MAP_FAILED == opd->pidm) {
        ifBLIN_QW0("mmap");
        EXOUT(EX_IOERR);
    }
    close(opd->pidd);
    opd->pidm->mark = AUWITH_MARK;
    w = snprintf(opd->pidm->pid, 8, "%d\n", getpid());
    if  (0 > w) {
        ifBLIN_QW0("snprintf pid");
        EXOUT(EX_IOERR);
    } else if (w > 7) {
        ifBLIN_QX0("snprintf pid big");
        ERROUT(EX_SOFTWARE, EOVERFLOW);
    }
    opd->pidm->flags = opd->flags;
    opd->pidm->ninp = opd->inpachsz;
    opd->pidm->nout = opd->outachsz;
    for (i = 0; i < opd->pidm->ninp; ++i) {
        for (o = 0; o < opd->pidm->nout; ++o) {
            cell = auwith_cella(opd->pidm, i, o);
            cell->inp = i;
            cell->out = o;
            cell->flags = opd->flags & (BLIN_MASK | AUWITH_CFAF);
    }   }
    opd->inpvol = (auwith_vol*)&opd->pidm->matrix[opd->pidm->nout * opd->pidm->ninp];
    opd->outvol = &opd->inpvol[opd->pidm->ninp];
out:
    ifBLIN_QX3("- %d", ex);
    return(ex);
}

int
/*****************************************************************************************************
 *****************************************************************************************************/
main(int argc, char *argv[]) {                                                                   /****
 *****************************************************************************************************
 *****************************************************************************************************/
    const char *plugin;
    struct kevent ev;
    auwith_opd *opd;
    auwith_ent ent;
    int ex = EX_OK;
    u_int i, o, j;

    blin_ctl(BLIN_CTL_FSET, BLIN_BIT0);
//  blin_ctl(BLIN_CTL_FSET, BLIN_MASK);
    if  (!(opd = calloc(1, sizeof(auwith_opd)))) {
        ifBLIN_QX0("Malloc failed #1");
        ERROUT(EX_OSERR, ENOMEM);
    }
    opd->flags = blin_internal_flags & BLIN_MASK;
    opd->pluginsz = AUWITH_PLUGINSZ;
    opd->inpachsz = AUWITH_NINP;
    opd->outachsz = AUWITH_NOUT;
    opd->nextplugin = 0;
    opd->pidf = AUWITH_PID;
    if  (!!(ex = parsins(opd, argc, argv))) {
        ifBLIN_QW0("parsins");
        goto out;
    }
    if  (!!(ex = plugins(opd))) {
        ifBLIN_QW0("plugins set");
        goto out;
    }
    if  (!(opd->inpach = calloc(opd->inpachsz, sizeof(auwith_io*)))) {
        ifBLIN_QX0("Malloc failed #2");
        ERROUT(EX_OSERR, ENOMEM);
    }
    if  (!(opd->outach = calloc(opd->outachsz, sizeof(auwith_io*)))) {
        ifBLIN_QX0("Malloc failed #3");
        ERROUT(EX_OSERR, ENOMEM);
    }
    if  (0 > (opd->kq = kqueue())) {
        ifBLIN_QW0("kqueue");
        EXOUT(EX_IOERR);
    }
    bzero(&ent, sizeof(ent));
    ent.flags = (opd->flags & BLIN_MASK) | (auwith_def_base.flags & ~BLIN_MASK);
    ent.plug = &auwith_def_base;
    while (!!(plugin = babolo_getargs(opd->bos))) {
        ifBLIN_QX5("plug=%s~", plugin);
        for (i = 0; i < opd->nextplugin; ++i) {
            ifBLIN_QX5("plug %u %s?", i, opd->plug[i]->pname);
            if  (  (opd->plug[i]->mark.mark = opd->plug[0]->mark.mark)
                && !strncmp(opd->plug[i]->pname, plugin, AUWITH_NAMESZ)
                ) break;
        }
        if  (i >= opd->nextplugin) {
            ifBLIN_QX0("Plugin %s not found", plugin);
            ERROUT(EX_USAGE, EPROTO);
        }
        if  (!i) break;
        if  (!!(ex = opd->plug[i]->parsin(opd))) {
            ifBLIN_QW0("Plugin %s operands not parsed", plugin);
            goto out;
    }   }
    for (j = 0, i = 0; j < opd->inpachsz; ++j) if (!!opd->inpach[j]) i = j + 1;
    opd->inpachsz = i;
    for (j = 0, o = 0; j < opd->outachsz; ++j) if (!!opd->outach[j]) o = j + 1;
    opd->outachsz = o;
    if  (!o || !i) {
        ifBLIN_QX0("NULL matrix IN %u x %u OUT", opd->inpachsz, opd->outachsz);
        ERROUT(EX_USAGE, ENXIO);
    }
    if  (!!(ex = createcontrolfile(opd))) {
        ifBLIN_QW0("createcontrolfile");
        goto out;
    }
    if  (!(opd->ref = calloc(i * o, sizeof(auwith_hc)))) {
        ifBLIN_QX0("Malloc failed #5");
        ERROUT(EX_OSERR, ENOMEM);
    }
    if  (!!plugin && !!(ex = opd->plug[0]->parsin(opd))) {
        ifBLIN_QW0("Plugin %s operands not parsed", plugin);
        goto out;
    }
    if  (opd->flags & AUWITH_DECONF) deconf(opd, NULL, stdout);
    auwith_Tinit(opd);
    if  (!!(ex = activ(opd, &ent, AUWITH_KIND_TIME))) {
        ifBLIN_QW0("base activ");
        goto out;
    }
    if  ((opd->pidm) && (opd->flags & AUWITH_LOOK)) auwith_matrix(opd->pidm, stdout);
    for (; !ex ;) {
        auwith_plug *p;

        auwith_Tmark(opd, 0);
        ex = auwith_event(opd, &ev);
        auwith_Tmark(opd, 1);
        if  (0 > ex) {
            ifBLIN_QW0("event");
            EXOUT(EX_IOERR);
        } else if (ex > 1) {
            ifBLIN_QX0("event=%d (?)", ex);
            ERROUT(EX_OSERR, EOVERFLOW);
        } else if (!ex) {
            p = ((auwith_ent*)ev.udata)->plug;
            ifBLIN_QX4("timeout %s", p->pname);
            if  (p->activ) {
                if  (!!(ex = p->activ(opd, ev.udata, AUWITH_KIND_TIME))) {
                    ifBLIN_QW0("activ %s at time", p->pname);
                    goto out;
            }   }
        } else {
            p = ((auwith_ent*)ev.udata)->plug;
            ifBLIN_QX4("action %s", p->pname);
            if  (p->activ) {
                if  (!!(ex = p->activ(opd, ev.udata, AUWITH_KIND_KQ))) {
                    ifBLIN_QW0("activ %s at kq", p->pname);
                    goto out;
    }   }   }   }
out:
    ifBLIN_QX3("- %d", ex);
    exit(ex);
}

/*****************************************************************************************************
 *****************************************************************************************************/
auwith_plug auwith_def_base =                                                                    /****
 *****************************************************************************************************
 *****************************************************************************************************/
{ "base"
, AUWITH_MARK
, parsin
, deconf
, activ
, NULL
, NULL
, NULL
, AUWITH_CAP_SPATIN
, 0
};
