/*-
 * Copyright (C)2014 @BABOLO http://www.babolo.ru/
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: auwith.h,v 1.1 2014/08/14 09:34:37 babolo Exp $
 */

#include <sys/endian.h>

#define AUWITH_MARK     (auwith_mark){{VMAJOR & 0x000000000000FFFFULL, 'A', 'U', 'W', 'I', 'T', 'H'}}
#define AUWITH_NFD      2                          /*    ( MIX  DEV)   */
#define AUWITH_MACH     8                          /*                   */
#define AUWITH_MWBY     2                          /*                    */
#define AUWITH_PID      "/var/run/auwithd.pid"
#define AUWITH_NAMESZ   8                          /*                              */
#define AUWITH_FASMUL   64
#define AUWITH_FASMIN   4
#define AUWITH_FASMAX   132

typedef struct auwith_bchain auwith_bchain;
typedef struct auwith_opd auwith_opd;
typedef struct auwith_ent auwith_ent;
typedef struct auwith_io auwith_io;
typedef struct auwith_times auwith_times;
typedef struct auwith_plug auwith_plug;
typedef u_int64_t auwith_time;
                        /*****************************/
typedef union {         /********* auwith_mark *******/
    struct {
        u_int64_t ver : 16
                , c0  : 8
                , c1  : 8
                , c2  : 8
                , c3  : 8
                , c4  : 8
                , c5  : 8
        ;
    };
    u_int64_t mark;
} auwith_mark;
                        /*****************************/
typedef struct {        /********* auwith_vol ********/
    int16_t         min;
    int16_t         max;
} auwith_vol;
                        /*****************************/
typedef struct {        /********* auwith_env ********/
    BLIN_flag       flags;
    int             dummy;
    union {
        int         aufd[AUWITH_NFD];
        struct {
            int     aumix;                      /*                       */
            int     audev;                      /*                         */
    };  };
    union {
        const char *fnm[AUWITH_NFD];
        struct {
            const char *mix;                    /*                             */
            const char *dev;                    /*                               */
    };  };
} auwith_env;
                        /*****************************/
typedef struct {        /********* auwith_buf ********/
    int             refcnt;
    int             size;                       /*                                         */
# if defined(__GNUC__) && __GNUC__ > 2
    int16_t         samples[];
# else
    int16_t         samples[0];
# endif
} auwith_buf;
                        /*****************************/
struct auwith_bchain {  /********* auwith_bchain *****/
    auwith_bchain  *next;
    auwith_buf     *buf;
};
                        /*****************************/
typedef struct {        /********* auwith_hc *********/
    auwith_bchain  *bchain;
    off_t           offset;
    int16_t         pobuf;
    int16_t         sample;
    char            ready;
} auwith_hc;
                        /*****************************/
typedef struct {        /********* auwith_sfaf *******/
    auwith_time     tick[4];
    u_int16_t       chainlen;
    int16_t         faser;
    char            doo;
} auwith_sfaf;
                        /*****************************/
typedef struct {        /********* auwith_cell *******/
    BLIN_flag       flags;
#   define       AUWITH_LEVEL  0x00FFFF
#   define       AUWITH_ALGOS  16
#   define       AUWITH_ALGO/* 0x1F0000 */ (0x1F << AUWITH_ALGOS)
#   define       AUWITH_CFAF   0x400000         /*!   cell,  _opd                 */
#   define       AUWITH_USED   0x800000
    u_int16_t       inp;
    u_int16_t       out;
    auwith_sfaf     f;
    auwith_sfaf     t;
} auwith_cell;
                        /*****************************/
typedef struct {        /********* auwith_pid ********/
    char            pid[8];
    auwith_mark     mark;
    BLIN_flag       flags;
    u_int16_t       ninp;
    u_int16_t       nout;
# if defined(__GNUC__) && __GNUC__ > 2
    auwith_cell     matrix[];
# else
    auwith_cell     matrix[0];
# endif
/*  auwith_vol      inpvol[.ninp]; */
/*  auwith_vol      outvol[.nout]; */
} auwith_pid;
                        /*****************************/
struct auwith_times {   /********* auwith_times ******/
    auwith_times   *next;
    void           *udata;
    auwith_time     time;
};
                        /*****************************/
typedef struct {        /********* auwith_interhand **/
    int           (*init)(auwith_opd *opd, auwith_cell *cell);
    int           (*fini)(auwith_opd *opd, auwith_cell *cell);
    int32_t       (*get)(auwith_opd *opd, auwith_cell *cell);
    void          (*take)(auwith_opd *opd, auwith_cell *cell);
    auwith_buf   *(*read)(auwith_opd *opd, auwith_cell *cell);
    int           (*write)(auwith_opd *opd, auwith_cell *cell, auwith_buf *buf);
} auwith_interhand;

# if defined(_BABOLO_PARSER_H_)
                        /*****************************/
struct auwith_plug {    /********* auwith_plug *******/
    char            pname[AUWITH_NAMESZ];       /*                                      */
    auwith_mark     mark;
    int           (*parsin)(auwith_opd *opd);
    void          (*deconf)(auwith_opd *opd, auwith_io *io, FILE *fo);
    int           (*activ)(auwith_opd *opd, auwith_ent *ent, int kind);
#   define  AUWITH_KIND_TIME   0                /* !!! NOT redefineable                           */
#   define  AUWITH_KIND_KQ     1                /* !!! NOT redefineable                           */
    int           (*passiv)(auwith_opd *opd, auwith_cell *cell, int kind);
#   define  AUWITH_KIND_WRITE  2
#   define  AUWITH_KIND_MIXEN  3
#   define  AUWITH_KIND_MIXDI  4
    auwith_buf   *(*reed)(auwith_opd *opd, auwith_cell *cell);
    int           (*rite)(auwith_opd *opd, auwith_cell *cell, auwith_buf *buf);
    BLIN_flag       flags;                      /*                                 */
#   define  AUWITH_CAP_SYNCNO  0x000000         /*                                     */
#   define  AUWITH_CAP_SYLAZY  0x000001         /*                                    */
#   define  AUWITH_CAP_SPATIN  0x000002         /*                                      */
#   define  AUWITH_CAP_STRICT  0x000003         /*                                    */
#   define  AUWITH_CAP_SMASK   0x000003         /*                              */
#   define  AUWITH_CAP_SKEW    0x000004         /*                         */
#   define  AUWITH_CAP_SFULL/* 0x000007 */ (AUWITH_CAP_SMASK | AUWITH_CAP_SKEW)
#   define  AUWITH_CAP_SFULLS  3                /* first (3 == AUWITH_CAP_SFULLS) bits - IMPOTANT */
#   define  AUWITH_CAP_INP     0x000010         /*                             */
#   define  AUWITH_CAP_OUT     0x000020         /*                             */
#   define  AUWITH_CAP_SWAPS   6                /*                                    */
#   define  AUWITH_CAP_SWAP /* 0x000FC0 */ (0x3F << AUWITH_CAP_SWAPS)
#   define  AUWITH_CAP_ALAW    0x001000         /*  G.711 A-law                                */
#   define  AUWITH_CAP_uLAW    0x002000         /*  G.711 u-law                                */
#   define  AUWITH_CAP_LOCAL   0x004000         /*                     */
#   define  AUWITH_CAP_MASK    0x00FFFF
    u_int32_t       xormask;                    /*                                                */
};
                        /*****************************/
struct auwith_ent {     /********* auwith_ent ********/
    BLIN_flag       flags;
/*          AUWITH_CAP_MASK    0x00FFFF          * ,   _plug,   _opd  */
    u_int32_t       xormask;                    /*                                                */
    auwith_plug    *plug;
    auwith_io      *in[AUWITH_MACH];
    auwith_io      *ou[AUWITH_MACH];
    ssize_t         block;                      /*    samples                         */
    u_int32_t       rate;                       /*    Hz                          */
    u_int16_t       bits;                       /*                             */
    u_int16_t       nchan;                      /*                               */
};
                        /*****************************/
struct auwith_io {      /********* auwith_io *********/
    BLIN_flag       flags;                      /*   _ent                              */
    u_int32_t       sock;
    auwith_ent     *ent;
};
                        /*****************************/
typedef struct {        /********* auwith_opt ********/
    BLIN_flag       flags;                      /*                                      */
#   define       AUWITH_MAST   0x002000         /*                         */
#   define       AUWITH_FAEX   0x004000         /*                    */
#   define       AUWITH_DIAL   0x010000         /*                                          */
#   define       AUWITH_LOOK   0x020000         /*!                                */
#   define       AUWITH_FASE   0x040000         /*                                */
#   define       AUWITH_uSEC   0x080000         /* us  ns  -T                            */
#   define       AUWITH_FILT   0x100000         /*    -T                             */
#   define       AUWITH_ILEV   0x200000         /*                                */
#   define       AUWITH_OLEV   0x400000         /*                               */
    int             pidd;                       /*   pid                   */
    auwith_time     tr;                         /*                                 */
    const char     *pidf;                       /*  pid                                   */
    auwith_pid     *pidm;                       /*    pid                    */
    babolo_opts    *bos;                        /*                                */
    const char     *T;                          /*                                                */
    auwith_vol     *inpvol;                     /*                             */
    auwith_vol     *outvol;                     /*                             */
    u_int32_t       sock;                       /*                 */
    int             Tfilter;
    auwith_env      env;
    char            tt;                         /*                          */
} auwith_opt;
                        /*****************************/
struct auwith_opd {     /********* auwith_opd ********/
    BLIN_flag       flags;                      /*                                      */
#   define       AUWITH_HARDTI 0x010000         /*                */
/*               AUWITH_LOOK   0x020000          *                                 */
#   define       AUWITH_DECONF 0x040000         /*  deconf()                                */
#   define       AUWITH_DUMPOP 0x080000         /*  dumpopts()                              */
#   define       AUWITH_LOCAL  0x100000         /*                     */
/*               AUWITH_CFAF   0x400000          *    cell                          */
    int             pidd;                       /*   pid                   */
    auwith_time     tr;                         /*                                 */
    const char     *pidf;                       /*  pid                                   */
    auwith_pid     *pidm;                       /*    pid                    */
    babolo_opts    *bos;                        /*                                */
    const char     *T;                          /*                          */
    auwith_times   *times;                      /*                                */
    auwith_plug   **plug;                       /*                */
    auwith_io     **inpach;                     /*             */
    auwith_io     **outach;                     /*            */
    auwith_vol     *inpvol;                     /*                             */
    auwith_vol     *outvol;                     /*                             */
    auwith_hc      *ref;                        /*      cell             */
    int             t;                          /*                      */
    int             kq;                         /*   kevent                          */
    u_int16_t       pluginsz;                   /*   *plugin                    */
    u_int16_t       nextplugin;                 /*   *plugin                         */
    u_int16_t       inpachsz;                   /*   *inpach                    */
    u_int16_t       outachsz;                   /*   *outach                    */
};
# endif /* _BABOLO_PARSER_H_ */

# if 0
typedef struct auvocsd_cn auvocsd_cn;
struct auvocsd_cn {     /*********************/ /*    ,                */
    auvocsd_cn *next;                           /*                                   */
    char       *member;                         /*   "port:vol"                         */
};
typedef struct {        /*********************/ /*                               */
    addrinfo   *ai;                             /* addrinfo                              */
    size_t      p;                              /*                           */
    size_t      pp;                             /*                       */
    int         sock;                           /*                                  */
    int         kq;                             /*   kevent                          */
    u_int16_t   port;                           /*                                   */
    void       *buf1;                           /* 1                               */
    void       *buf2;                           /* 2                               */
    u_int32_t   flags;
#   define      AUWITH_BUF2 0x000200            /* 2                                   */
} auvocsd_bf;
typedef struct {        /*********************/ /*                       */
    char       *cport;                          /*                             */
    int32_t     buf;                            /*    options->bufs[]             */
    u_int32_t   level;                          /*                                */
    u_int16_t   port;                           /*                                    */
} auvocsd_ar;
typedef struct {        /*********************/ /*              */
    addrinfo   *ai;                             /*   addrinfo             */
    addrinfo  **ae;                             /*   addrinfo              */
    char       *tport;                          /*                       */
    int        *sock;                           /*                        */
    u_int16_t   nsock;                          /*                    */
    u_int16_t   port;                           /*                              */
} auwithd_cn;
typedef struct {
    u_int16_t    freq;                          /*   , Hz                   */
    u_int16_t    ampli;                         /*     level  */
} auvocsd_tsin;
typedef struct {        /*********************/ /*     pid                     */
    BLIN_flag    flags;                         /*                                           */
#   define       AUWITH_TFLAG  0x0000FF         /* -T                                         */
#   define       AUWITH_TEST   0x000100         /*                 */
#   define       AUWITH_ABARIX 0x000200         /*  G.711 A-low                                */
#   define       AUWITH_uBARIX 0x000400         /*  G.711 u-low                                */
#   define       AUWITH_FILE   0x000800         /*                                 */
#   define       AUWITH_LOOKFO 0x001000         /*                      */
#   define       AUWITH_LLEVEL 0x002000         /* aulook                      */
#   define       AUWITH_SWABY  0x004000         /* swap bytes                                     */
#   define       AUWITH_SWABI  0x008000         /* swap bits                                      */
#   define       AUWITH_NOWITH 0x010000         /*    auwith                           */
#   define       AUWITH_NOWITH 0x020000         /*    auvocs                           */
    pid_t        pid;                           /* PID                                            */
    u_int16_t    cbufl;                         /*    UDP                */
    u_char       cbips;                         /*      UDP           */
    u_char       cbyps;                         /*      UDP          */
    int          crte;                          /*    UDP                  */
    int          rate;                          /*    Hz                          */
    u_char       bitps;                         /*                             */
    u_char       nchan;                         /*      aulook          */
    u_char       bytps;                         /*                            */
    u_int16_t    bufln;                         /*    snd                */
    u_int32_t    level;                         /*                               */
    int          sinn;                          /*                  */
    auvocsd_tsin tsin[AUWITH_MACH];             /*  (Hz)          */
} auwith_comm;
typedef struct {
    size_t       curr;
    size_t       phase;
    u_int32_t    uvlc;
    u_int32_t    ovlc;
    u_int16_t    port;
} auvocsd_stat;
typedef struct {
    char         pid[8];
    union {
        u_int64_t   mark;
        struct {
#if _BYTE_ORDER == _LITTLE_ENDIAN
            u_int16_t  major;
            char       kind[6];
#else /* _BYTE_ORDER */
            char       kind[6];
            u_int16_t  major;
#endif /* _BYTE_ORDER */
    };  };
    auwith_comm  c;
    int auspl;                                  /*                                    */
    int mi[AUWITH_MACH];                        /*                       */
    int ma[AUWITH_MACH];                        /*                      */
    union {
        struct {        /*********************/ /*  auvocsd                               */
            u_char       bufn;                  /*                 */
            u_char       nin[AUWITH_MACH];      /*       */
            auvocsd_stat stat[];                /*                           */
        };
        struct {        /*********************/ /*  auwithd                               */
            u_int32_t    voice[AUWITH_MACH];    /*                             */
    };  };
} auvocsd_pid;
typedef struct {        /*********************/ /*                              */
    auwith_comm  c;
    auvocsd_pid *pid;                           /*    pid                    */
    char        *pidf;                          /*  pid                                   */
    int          pidd;                          /*   pid                   */
    union {struct {
            int  aumix;                         /*                       */
            int  audev;                         /*                         */
        };
        int      aufd[AUWITH_NFD];
    };
    union {struct {
            char *mix;                          /*                             */
            char *dev;                          /*                               */
        };
        char     *fnm[AUWITH_NFD];
    };
} auwith_cf;
typedef struct {        /*********************/ /*                                          */
    auwith_cf    cf;                            /*                              */
    babolo_opt *opt;                            /*           */
    struct timespec tr;                         /*                                  */
    union {
        struct {        /*********************/ /*  auvocsd                                 */
            char        *host;                  /*   bind                                 */
            auwith_kev  *kev;                   /*  kevent                          */
            char        *xlog;                  /*      BARIX           */
            int          xlod;                  /*     BARIX                */
            int          kq;                    /*   kevent                          */
            auvocsd_ar  *combo[AUWITH_MACH];    /*             */
            auvocsd_cn  *ochan[AUWITH_MACH];    /*         */
            auvocsd_bf  *bufs;                  /*                */
            u_int16_t    ichn;                  /*                        */
            u_char       nin[AUWITH_MACH];      /*       */
            u_int16_t    dummy0;
        };
        struct {        /*********************/ /*  auwithd  aulook                        */
            struct timespec ts;                 /*                                   */
            struct timespec tw;                 /*                                     */
            union {
                struct {/*********************/ /*  auwithd                                 */
                    auwith_bufc bufc[AUWITH_MACH];/*                     */
                    size_t      p;              /*                          */
                    u_int16_t   port;           /*                                     */
                    auwithd_cn  channels[AUWITH_MACH];/*     */
                    schar       linin;          /*                          */
                    u_char      dummy1;
                };
                struct {/*********************/ /*  aulook                                  */
                    int  auspl;                 /*                                    */
                    int  d;                     /*     b[]                */
                    u_int16_t b[16];            /*                 */
    };  };  };  };
} auvocs_cf;
int    auwith_show   __P((auwith_cf *cf, int full));
int    auwith_do     __P((auwith_cf *cf, schar linin));
int    auwith_ana    __P((auwith_cf *cf, schar linin));
int    auwith_ctlf   __P((auwith_cf *cf, u_int64_t kind));
double auwith_sin    __P((auwith_cf *cf, int chan, int rate));
int    auwith_parsin __P((auvocs_cf *options, int argc, char *argv[], char **envp));
int    auvocs_init   __P((auvocs_cf *options));
long   auvocs_get    __P((auvocs_cf *options, int32_t bufn));
int    auvocs_next   __P((auvocs_cf *options));
int    auwith_init   __P((auvocs_cf *options));
int    auwith_put    __P((auvocs_cf *options, int chan, int32_t v));
int    auwith_next   __P((auvocs_cf *options));
# endif

# if defined(_BABOLO_PARSER_H_)
extern auwith_plug auwith_def_base;
extern auwith_plug auwith_def_sine;
extern auwith_plug auwith_def_dspo;
extern auwith_plug auwith_def_dspi;
extern auwith_plug auwith_def_udpo;
extern auwith_plug auwith_def_udpi;
extern auwith_plug auwith_def_file;
#  if defined(__GNUC__) && __GNUC__ > 2
//extern auwith_plug *auwith_allplug[];
extern auwith_interhand auwith_handlers[];
#  else
//extern auwith_plug *auwith_allplug[0];
extern auwith_interhand auwith_handlers[0];
#  endif
# endif /* _BABOLO_PARSER_H_ */

void         auwith_caproc   __P((auwith_plug *plug, FILE *fo));
void         auwith_capand   __P((BLIN_flag flags, FILE *fo, int kind));
int          auwith_show     __P((auwith_env *env, FILE *fo, u_int full));
                                      /* full: 1 /dev/mixer, 2 /dev/dsp, 4 mixer matrix, 8 ext info */
int          auwith_ana      __P((auwith_env *env, FILE *fo));
void         auwith_matrix   __P((auwith_pid *pidm, FILE *fo));
auwith_cell *auwith_cella    __P((auwith_pid *pidm, u_int i, u_int o));
char        *auwith_todev    __P((const char *o, char *d));
int          auwith_totime   __P((const char *o, struct timespec *tr));
int          auwith_plumb    __P((auwith_opd *opd, auwith_io *io, int kind));
                                      /* kind: 0 input, 1 output */
u_int        auwith_handle   __P((auwith_opd *opd, u_int i, u_int o));
auwith_time  auwith_future   __P((auwith_time time));
auwith_time  auwith_bufuture __P((auwith_ent *ent, int64_t v));
auwith_time  auwith_fifuture __P((auwith_ent *ent));
void         auwith_Tinit    __P((auwith_opd *opd));
void         auwith_Tmark    __P((auwith_opd *opd, int32_t c));
auwith_buf  *auwith_prepare  __P((auwith_opd *opd, auwith_ent *ent));
int          auwith_postpare __P((auwith_opd *opd, auwith_ent *ent, void *inpb, int p));
auwith_buf  *auwith_creabuf  __P((size_t size));
void         auwith_relebuf  __P((auwith_buf *buf));
void         auwith_takebuf  __P((auwith_buf *buf));
auwith_hc   *auwith_cadd     __P((auwith_opd *opd, auwith_cell *cell));
int          auwith_sfafer   __P((auwith_sfaf *sfaf));

# if defined(_SYS_EVENT_H_)
int          auwith_queue  __P((auwith_opd *opd, struct kevent *ev, void *udata, auwith_time time));
int          auwith_event  __P((auwith_opd *opd, struct kevent *ev));
# endif /* _SYS_EVENT_H_ */
