/*-
 * Copyright (C)2014 @BABOLO http://www.babolo.ru/
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static const char copyright[] = "\
@(#)Copyright (C)2014 @BABOLO http://www.babolo.ru/\n\
@(#)All rights reserved.\n\
@(#)V.M "VERS;
static const char rcsid[] = "$Id: dspi.c,v 1.1 2014/08/14 09:34:37 babolo Exp $";
#endif /* not lint */

#define BLIN_COMPAT  3
#define Bpars_COMPAT 3

#include <sys/types.h>
#include <sys/soundcard.h>
#include <sys/event.h>
#include <sys/time.h>
#include <sysexits.h>
#include <strings.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>
#include <errno.h>
#include <err.h>
#include <babolo/BLINflag.h>
#include <babolo/parser.h>
#include "auwith_int.h"
                        /*****************************/
typedef struct {        /********* auwith_ent_dspi ***/
    auwith_ent      common;
    auwith_time     base;
    auwith_env      env;
    ssize_t         inpmem;                     /*                 */
    void           *inpb;
} auwith_ent_dspi;
                        /*****************************/
typedef struct {        /********* auwith_io_dspi ****/
    auwith_io       common;
    u_int16_t       chan;
} auwith_io_dspi;

static auwith_time sta = 0;

static void
/*****************************************************************************************************/
usage(void) {                                                                                    /****
 *****************************************************************************************************/
    fprintf( stderr
           , "auwithd dspi V%u @BABOLO " DATE "\n"
             "Usage: auwithd [-options] dspi [-options]\n"
             "   -a dev          DSP to work with, default %s\n"
             "   -c num          channels number\n"
             "   -D              hard timing debug\n"
             "   -E -X XX        use XX as env name for opt's arg\n"
             "   -i in           matrix's input number\n"
             "   -L              local phase count\n"
             "   -q              silent\n"
             "   -r freq         samples rate in Hz (default %u)\n"
             "   -s num          block size in samples (default %u)\n"
             "   -v[v[v...]]     verbose\n"
             "   -z TEXT         parce TEXT as arguments continuation\n"
             "   -Z TEXT         parce TEXT as arguments continuation with env substitution\n"
             "   -= mode         internal diagnostic\n"
             "   -?              this text\n"
           , auwith_def_dspi.mark.ver
           , AUWITH_DEV
           , AUWITH_RATE
           , AUWITH_BLOCK
           )
    ;
}

static int
/*****************************************************************************************************/
parsin(auwith_opd *opd) {                                                                        /****
 *****************************************************************************************************/
    char *flar = "a:c:Di:Lqr:s:v=:?";
    int iblsize, samplewidth;
    struct snd_size blocksz;
    auwith_ent_dspi *ent;
    int c, ex = EX_OK;
    audio_buf_info bi;
    u_int16_t inpp;

    ifBLIN_QX3("+");
    if  (!opd) {
        ifBLIN_QX0("no opd");
        ERROUT(EX_DATAERR, EFAULT);
    }
#   define blin_internal_flags (opd->flags & BLIN_MASK)
    if  (!opd->bos) {
        ifBLIN_QX0("no bos");
        ERROUT(EX_DATAERR, EFAULT);
    }
    if  (!(ent = calloc(1, sizeof(auwith_ent_dspi)))) {
        ifBLIN_QX0("Malloc failed #1");
        ERROUT(EX_OSERR, ENOMEM);
    }
    ent->common.flags = (opd->flags & ~AUWITH_CAP_MASK) | (auwith_def_dspo.flags & AUWITH_CAP_MASK);
    ent->common.plug = &auwith_def_dspi;
    ent->common.rate = AUWITH_RATE;
    ent->common.block = AUWITH_BLOCK;
    ent->env.dev = AUWITH_DEV;
    ent->common.bits = AUWITH_BITS;
    if  (!!babolo_chopts(opd->bos, 0, flar)) {
        ifBLIN_QW0("chopts");
        EXOUT(EX_SOFTWARE);
    }
    ex = EX_OK;
    if  (opd->flags & AUWITH_DUMPOP) babolo_dumpopts(opd->bos);
    while (!!(c = babolo_getopts(opd->bos))) {
        switch (c) {
        case 'a': if  (!(ent->env.dev = auwith_todev(babolo_getoptsarg(opd->bos), AUWITH_DEV)))
                      goto out;
                ; break;
        /****************************************************************************/
        case 'b': ent->common.bits = babolo_getoptunum(opd->bos, 0); /* not reached */
                ; break;                                             /* not reached */
        /****************************************************************************/
        case 'c': if  (!!ent->common.nchan) {
                      ifBLIN_QX0("redefine nchan");
                      ERROUT(EX_USAGE, ENOTEMPTY);
                  }
                ; ent->common.nchan = babolo_getoptunum(opd->bos, 0);
                ; inpp = 0;
                ; break;
        case 'D': ent->common.flags |= AUWITH_HARDTI;
                ; break;
        case 'i': if  (inpp >= ent->common.nchan) {
                      ifBLIN_QX0("chan %u >= %u", inpp, ent->common.nchan);
                      EXOUT(EX_USAGE);
                  }
                ; if  (!(ent->common.in[inpp] = calloc(1, sizeof(auwith_io_dspi)))) {
                       ifBLIN_QX0("Malloc failed #2");
                       ERROUT(EX_OSERR, ENOMEM);
                  }
                ; ent->common.in[inpp]->flags = ent->common.flags & ~AUWITH_CAP_OUT;
                ; ent->common.in[inpp]->ent = (auwith_ent*)ent;
                ; ((auwith_io_dspi*)ent->common.in[inpp])->chan = inpp;
                ; ent->common.in[inpp]->sock = babolo_getoptunum(opd->bos, 0);
                ; if  (ent->common.in[inpp]->sock >= opd->inpachsz) {
                      ifBLIN_QX0( "-i[%u] must be < %u, now %u"
                                , inpp, opd->inpachsz, ent->common.in[inpp]->sock
                                );
                      ERROUT(EX_USAGE, EINVAL);
                  }
                ; ++inpp;
                ; break;
        case 'L': ent->common.flags ^= AUWITH_CFAF;
                ; break;
        case 'q': BLIN_QUIET(ent->common.flags);
                ; break;
        case 'r': ent->common.rate = babolo_getoptunum(opd->bos, 0);
                ; break;
        case 's': ent->common.block = babolo_getoptunum(opd->bos, 0);
                ; break;
        case 'v': BLIN_VERBOSE(ent->common.flags);
                ; break;
        case '=': switch(babolo_getoptsarg(opd->bos)[0]) {
                  case '0': blin_ctl(BLIN_CTL_DUMP);           exit(EX_OK);
                  case '1': babolo_optest(flar);               exit(EX_OK);
                  case '2': babolo_optext(opd->bos);           exit(EX_OK);
                  case '3': babolo_dumpopts(opd->bos);         exit(EX_OK);
                  }
                ; break;
        case '?': usage();
                ; break;
        default : ifBLIN_QX0("Illegal flag -%c", opd->bos->c);
                ; usage();
                ; ERROUT(EX_USAGE, EINVAL);
    }   }
    ent->env.flags = ent->common.flags & BLIN_MASK;
    samplewidth = (ent->common.bits + 7) / 8;
    iblsize = samplewidth * ent->common.nchan * ent->common.block;
    if  (!ent->common.bits || (ent->common.bits > 16)) {
        ifBLIN_QX0("-b must be >0 and <=16, now %u", ent->common.bits);
        ERROUT(EX_USAGE, EINVAL);
    }
    if  (!ent->common.rate || (ent->common.rate > 48000)) {
        ifBLIN_QX0("-r must be >0 and <=48000, now %u", ent->common.rate);
        ERROUT(EX_USAGE, EINVAL);
    }
    if  (!ent->common.block || (ent->common.block > 16384)) {
        ifBLIN_QX0("-s must be >0 and <=16384, now %u", ent->common.block);
        ERROUT(EX_USAGE, EINVAL);
    }
    if  (!ent->env.dev) {
        ifBLIN_QX0("no dsp selected");
        ERROUT(EX_USAGE, EINVAL);
    }
    if  (0 > (ent->env.audev = open(ent->env.dev, O_RDWR))) {
        ifBLIN_QW0("open audio dev");
        EXOUT(EX_IOERR);
    }
    c = ent->common.bits;
    if  (0 > ioctl(ent->env.audev, SNDCTL_DSP_SETFMT, &c)) {
        ifBLIN_QW0("audio bits set");
        EXOUT(EX_IOERR);
    }
    if  (c != ent->common.bits) ifBLIN_QX1("audio bits %u != %u", c, ent->common.bits);
    c = ent->common.nchan;
    if  (0 > ioctl(ent->env.audev, SNDCTL_DSP_CHANNELS, &c)) {
        ifBLIN_QW0("audio chan set");
        EXOUT(EX_IOERR);
    }
    if  (c != ent->common.nchan) ifBLIN_QX1("audio chan %u != %u", c, ent->common.nchan);
    c = ent->common.rate;
    if  (0 > ioctl(ent->env.audev, SNDCTL_DSP_SPEED, &c)) {
        ifBLIN_QW0("audio rate set");
        EXOUT(EX_IOERR);
    }
    if  (c != (int)ent->common.rate) ifBLIN_QX1("audio rate %u != %u", c, ent->common.rate);
    blocksz.rec_size = iblsize;
    blocksz.play_size = iblsize;
    if  (0 > ioctl(ent->env.audev, AIOSSIZE, &blocksz)) {
        ifBLIN_QW0("audio block size set");
        EXOUT(EX_IOERR);
    }
    bzero(&bi, sizeof(bi));
    if  (0 > ioctl(ent->env.audev, SNDCTL_DSP_GETISPACE, &bi)) {
        ifBLIN_QW0("audio rec mem space get");
        EXOUT(EX_IOERR);
    }
    ent->inpmem = bi.fragstotal * bi.fragsize;
    if  (!!bi.bytes) ifBLIN_QX1("bytes %u<>0", bi.bytes);
    for (inpp = 0; !ex && (inpp < ent->common.nchan); ++inpp) {
        if  (!!(ex = auwith_plumb(opd, (auwith_io*)ent->common.in[inpp], 0))) {
            ifBLIN_QX0("auwith_plumbi");
            goto out;
    }   }
    c = PCM_ENABLE_INPUT;
    if  (0 > ioctl(ent->env.audev, SNDCTL_DSP_SETTRIGGER, &c)) {
        ifBLIN_QW0("audio set trigger");
        EXOUT(EX_IOERR);
    }
    if  (!(ent->inpb = malloc(iblsize))) {
        ifBLIN_QX0("Malloc failed #3");
        ERROUT(EX_OSERR, ENOMEM);
    }
    ent->base = auwith_fifuture(&ent->common);
    ex = auwith_queue(opd, NULL, ent, ent->base);
    if  (!!ex) {
        ifBLIN_QW0("queue");
        goto out;
    }
out:
    if  (!!ex && !!ent) {
        for (inpp = 0; inpp < ent->common.nchan; ++inpp) {
            if  (!!ent->common.in[inpp]) free(ent->common.in[inpp]);
        }
        free(ent->inpb);
        free(ent);
    }
    ifBLIN_QX3("- %d", ex);
    return(ex);
#   undef blin_internal_flags
}

static void
/*****************************************************************************************************/
deconf(auwith_opd *opd, auwith_io *io, FILE *fo) {                                               /****
 *****************************************************************************************************/
    auwith_io_dspi *cf = (auwith_io_dspi*)io;
    auwith_ent_dspi *ent;

#   define blin_internal_flags (io->flags & BLIN_MASK)
    ent = (auwith_ent_dspi*)cf->common.ent;
    fprintf( fo, "[%u] dspi %uHz, %usamples, chan %u, %s(%d)\n"
           , cf->common.sock
           , ent->common.rate
           , ent->common.block
           , cf->chan
           , ent->env.dev
           , ent->env.audev
           );
    auwith_show(&ent->env, fo, 6);
    fprintf(fo, "    recmem %u\n", ent->inpmem);
    fprintf(fo, "    ");
    auwith_capand(opd->flags, fo, 0);
    fprintf(fo, " ");
    auwith_capand(io->ent->flags, fo, 1);
    fprintf(fo, " ");
    auwith_capand(io->flags, fo, 1);
    fprintf(fo, "\n");
#   undef blin_internal_flags
}

static int
/*****************************************************************************************************/
activ(auwith_opd *opd, auwith_ent *ent, int kind) {                                              /****
 *****************************************************************************************************/
    int iblsize, samplewidth, p, ex = EX_OK;
    auwith_ent_dspi *cf = (auwith_ent_dspi*)ent;
    audio_buf_info bi;
    auwith_time tt;
    int64_t v;

#   define blin_internal_flags (ent->flags & BLIN_MASK)
    ifBLIN_QX3("+ %d", kind);
    if  (kind != AUWITH_KIND_TIME) {
        ifBLIN_QX0("kind = %d", kind);
        ERROUT(EX_SOFTWARE, EINVAL);
    }
    samplewidth = (ent->bits + 7) / 8;
    iblsize = samplewidth * ent->nchan * ent->block;
    bzero(&bi, sizeof(bi));
    if  (0 > ioctl(cf->env.audev, SNDCTL_DSP_GETISPACE, &bi)) {
        ifBLIN_QW0("audio play mem space get");
        EXOUT(EX_IOERR);
    }
    if  (bi.bytes < iblsize) {
        v = iblsize - bi.bytes;
        if  (ent->flags & AUWITH_HARDTI) {
            tt = auwith_future(0);
            fprintf( stderr, "%u %u %u -- %lld %llu --\n"
                   , cf->inpmem, cf->inpmem - bi.bytes, iblsize, v, (tt - sta) / 1000
                   );
            sta = tt;
        }
        if  (!!(ex = auwith_queue(opd, NULL, ent, auwith_bufuture(ent, v)))) {
            ifBLIN_QW0("queue");
        }
        goto out;
    }
    auwith_Tmark(opd, 6);
    p = read(cf->env.audev, cf->inpb, iblsize);
    auwith_Tmark(opd, 7);
    if  (0 > p) {
        ifBLIN_QW0("read");
        EXOUT(EX_IOERR);
    } else if (p > iblsize) {
        ifBLIN_QX0("read %d > %d", p, iblsize);
        ERROUT(EX_IOERR, EIO);
    } else if (p < iblsize) {
        ifBLIN_QX1("read %d < %d", p, iblsize);
    }
    if  (!!(ex = auwith_postpare(opd, ent, cf->inpb, p))) {
        ifBLIN_QW0("auwith_postpare");
        goto out;
    }
    bzero(&bi, sizeof(bi));
    if  (0 > ioctl(cf->env.audev, SNDCTL_DSP_GETISPACE, &bi)) {
        ifBLIN_QW0("audio play mem space get");
        EXOUT(EX_IOERR);
    }
    v = 2 * iblsize - bi.bytes;
    if  (v < 0) v = v / 4 - 1;
    if  (v > 0) v = v / 4 + 1;
    if  (v < -iblsize / 16) v = -iblsize / 16;
    if  (v > iblsize / 16) v = iblsize / 16;
    v += iblsize;
    cf->base = auwith_bufuture(ent, v);
    if  (ent->flags & AUWITH_HARDTI) {
        tt = auwith_future(0);
        fprintf( stderr, "%u %u %u . %d %lld %llu\n"
               , cf->inpmem, cf->inpmem - bi.bytes, iblsize, p, v, (tt - sta) / 1000
               );
        sta = tt;
    }
    if  (!!(ex = auwith_queue(opd, NULL, ent, cf->base))) {
        ifBLIN_QW0("queue");
    }
out:
    ifBLIN_QX3("- %d", ex);
    return(ex);
#   undef blin_internal_flags
}

/*****************************************************************************************************
 *****************************************************************************************************/
auwith_plug auwith_def_dspi =                                                                    /****
 *****************************************************************************************************
 *****************************************************************************************************/
{ "dspi"
, AUWITH_MARK
, parsin
, deconf
, activ
, NULL
, NULL
, NULL
, AUWITH_CAP_INP | AUWITH_CAP_STRICT | AUWITH_CAP_SKEW
, 0
};
