/*-
 * Copyright (C)2014 @BABOLO http://www.babolo.ru/
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static const char copyright[] = "\
@(#)Copyright (C)2014 @BABOLO http://www.babolo.ru/\n\
@(#)All rights reserved.\n\
@(#)V.M "VERS;
static const char rcsid[] = "$Id: handle.c,v 1.1 2014/08/14 09:34:37 babolo Exp $";
#endif /* not lint */

#define BLIN_COMPAT  3
#define Bpars_COMPAT 3

#include <sys/types.h>
#include <sysexits.h>
#include <strings.h>
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <babolo/BLINflag.h>
#include <babolo/parser.h>
#include "auwith_int.h"

static int bchaincnt = 0;

static const u_int
/*****************************************************************************************************/
handle[64] =                                                                                     /****
 *****************************************************************************************************/
{     01, 01, 01, 01,  02, 02, 02, 02 /* i0 */
,     01, 01, 01, 01,  02, 02, 02, 02 /* i1 */
,     01, 01, 01, 01,  02, 02, 02, 02 /* i2 */
,     01, 01, 01, 01,  02, 02, 02, 02 /* i3 */

,     01, 01, 01, 01,  02, 02, 02, 02 /* i4 */
,     01, 01, 01, 01,  02, 02, 02, 02 /* i5 */
,     01, 01, 01, 01,  02, 02, 02, 02 /* i6 */
,     01, 01, 01, 01,  02, 02, 02, 02 /* i7 */
}; /* o0  o1  o2  o3   o4  o5  o6  o7       */

u_int
/*****************************************************************************************************
 *****************************************************************************************************/
auwith_handle(auwith_opd *opd, u_int i, u_int o) {                                               /****
 *****************************************************************************************************
 *****************************************************************************************************/
    return(  handle[ ( (opd->inpach[i]->ent->plug->flags & AUWITH_CAP_SFULL) << AUWITH_CAP_SFULLS)
                   | (opd->outach[o]->ent->plug->flags & AUWITH_CAP_SFULL)
                   ]
          << AUWITH_ALGOS
          );
}

static void
/*****************************************************************************************************/
mark(auwith_cell *cell, int kind) {                                                              /****
 *****************************************************************************************************/
#   define blin_internal_flags (cell->flags & BLIN_MASK)
    auwith_time now;

    now = auwith_future(0);
    switch (kind) {
    case 1: if  (65535 == cell->f.chainlen) {
                ifBLIN_QX0("inc chainlen when 65535");
            } else {
                ++cell->f.chainlen;
            }
          ; if  (cell->f.tick[2] < cell->f.tick[0]) cell->f.tick[2] = now; else cell->f.tick[0] = now;
          ; break;
    case 2: if  (!cell->f.chainlen) {
                ifBLIN_QX0("dec chainlen when 0");
            } else {
                --cell->f.chainlen;
            }
          ; if  (cell->f.tick[3] < cell->f.tick[1]) cell->f.tick[3] = now; else cell->f.tick[1] = now;
          ; break;
    case 3: if  (cell->f.tick[2] < cell->f.tick[0]) cell->f.tick[2] = now; else cell->f.tick[0] = now;
          ; if  (cell->f.tick[3] < cell->f.tick[1]) cell->f.tick[3] = now; else cell->f.tick[1] = now;
          ; break;
    }
#   undef blin_internal_flags
}

static auwith_bchain *
/*****************************************************************************************************/
creabchain(size_t size) {                                                                        /****
 *****************************************************************************************************/
    auwith_bchain *nchain = NULL;

    ifBLIN_QX4("+ %u", size);
    if  (!(nchain = malloc(sizeof(auwith_bchain)))) {
        ifBLIN_QX0("Malloc failed");
        errno = ENOMEM;
    } else if (!(nchain->buf = auwith_creabuf(size))) {
        ifBLIN_QW0("auwith_creabuf");
    } else {
        nchain->next = NULL;
    }
    ifBLIN_QX4("- %08X->%08X bchain %d", nchain, nchain->next, ++bchaincnt);
    return(nchain);
}

static auwith_bchain *
/*****************************************************************************************************/
catchbchain(auwith_buf *buf) {                                                                   /****
 *****************************************************************************************************/
    auwith_bchain *nchain = NULL;

    ifBLIN_QX4("+ %08X", buf);
    if  (!buf) {
    } else if (!(nchain = malloc(sizeof(auwith_bchain)))) {
        ifBLIN_QX0("Malloc failed");
        errno = ENOMEM;
    } else {
        nchain->next = NULL;
        nchain->buf = buf;
    }
    ifBLIN_QX4("- %08X->%08X bchain %d", nchain, nchain->next, ++bchaincnt);
    return(nchain);
}

static auwith_bchain *
/*****************************************************************************************************/
takebchain(auwith_buf *buf) {                                                                    /****
 *****************************************************************************************************/
    auwith_bchain *nchain = NULL;

    ifBLIN_QX4("+ %08X", buf);
    if  (!buf) {
    } else if (!(nchain = malloc(sizeof(auwith_bchain)))) {
        ifBLIN_QX0("Malloc failed");
        errno = ENOMEM;
    } else {
        nchain->next = NULL;
        nchain->buf = buf;
        auwith_takebuf(nchain->buf);
    }
    ifBLIN_QX4("- %08X->%08X bchain %d", nchain, nchain->next, ++bchaincnt);
    return(nchain);
}

static void
/*****************************************************************************************************/
relebchain(auwith_bchain **bchain) {                                                             /****
 *****************************************************************************************************/
    auwith_bchain *nchain;

    ifBLIN_QX4("+ %08X->%08X", bchain, *bchain);
    if  (!!*bchain) {
        nchain = (*bchain)->next;
        auwith_relebuf((*bchain)->buf);
        free(*bchain);
        *bchain = nchain;
    }
    ifBLIN_QX4("- ->%08X bchain %d", *bchain, --bchaincnt);
}

static int
/*****************************************************************************************************/
init01(auwith_opd *opd, auwith_cell *cell) {                                                     /****
 *****************************************************************************************************/
#   define blin_internal_flags (opd->flags & BLIN_MASK)
    int ex = EX_OK;

    ifBLIN_QX4("+ %u->%u", cell->inp, cell->out);
    cell->flags = (opd->flags & BLIN_MASK) | (cell->flags & ~BLIN_MASK);
    ifBLIN_QX4("- %d", ex);
    return(ex);
#   undef blin_internal_flags
}

static int
/*****************************************************************************************************/
init02(auwith_opd *opd, auwith_cell *cell) {                                                     /****
 *****************************************************************************************************/
#   define blin_internal_flags (opd->flags & BLIN_MASK)
    int ex = EX_OK;

    ifBLIN_QX4("+ %u->%u", cell->inp, cell->out);
    if  (!!(ex = init01(opd, cell))) {
        ifBLIN_QW0("init[02]");
        goto out;
    }
    auwith_cadd(opd, cell)->bchain = creabchain(opd->inpach[cell->inp]->ent->block * AUWITH_MWBY);            /*@*/
    mark(cell, 1);
out:
    ifBLIN_QX4("- %d", ex);
    return(ex);
#   undef blin_internal_flags
}

static int
/*****************************************************************************************************/
fini01(auwith_opd *opd, auwith_cell *cell) {                                                     /****
 *****************************************************************************************************/
#   define blin_internal_flags (opd->flags & BLIN_MASK)
    int ex = EX_OK;

    ifBLIN_QX4("+ %u->%u", cell->inp, cell->out);
    relebchain(&auwith_cadd(opd, cell)->bchain);                                                              /*@*/
    bzero(auwith_cadd(opd, cell), sizeof(auwith_hc));
    ifBLIN_QX4("- %d", ex);
    return(ex);
#   undef blin_internal_flags
}

static int
rd01(auwith_opd *opd, auwith_cell *cell) {                                                     /****
 *****************************************************************************************************/
#   define blin_internal_flags (cell->flags & BLIN_MASK)
    auwith_buf *buf = NULL;
    auwith_bchain *bchain;
    int ex = EX_OK;

    ifBLIN_QX4("+ %u->%u", cell->inp, cell->out);
    if  (!auwith_cadd(opd, cell)->bchain && !!opd->inpach[cell->inp]->ent->plug->reed) {
        errno = 0;
/*@*/   buf = opd->inpach[cell->inp]->ent->plug->reed(opd, cell);
        if  (!buf && !!errno) {
            ifBLIN_QW0("->reed[%u (%u)]", cell->inp, cell->out);
            EXOUT(EX_IOERR);
        }
/*@*/   bchain = catchbchain(buf);
        buf = NULL;
        if  (!!bchain) {
            auwith_cadd(opd, cell)->bchain = bchain;
            mark(cell, 1);
   }   }
out:
    ifBLIN_QX4("- %d", ex);
    return(ex);
#   undef blin_internal_flags
}

static auwith_buf *
/*****************************************************************************************************/
read01(auwith_opd *opd, auwith_cell *cell) {                                                     /****
 *****************************************************************************************************/
#   define blin_internal_flags (cell->flags & BLIN_MASK)
    auwith_buf *buf = NULL;

    ifBLIN_QX4("+ %u->%u", cell->inp, cell->out);
    if  (!!rd01(opd, cell)) {
        ifBLIN_QW0("rd01");
        goto out;
    }
    if  (!!auwith_cadd(opd, cell)->bchain) {
/*@*/   buf = auwith_cadd(opd, cell)->bchain->buf;
/*@*/   relebchain(&auwith_cadd(opd, cell)->bchain);
        mark(cell, 2);
    }
    auwith_cadd(opd, cell)->pobuf = 0;
    auwith_cadd(opd, cell)->ready = 0;
out:
    ifBLIN_QX4("- %s", buf ? "OK" : "NULL");
    return(buf);                                                                                  /*@*/
#   undef blin_internal_flags
}

static int32_t
/*****************************************************************************************************/
get01(auwith_opd *opd, auwith_cell *cell) {                                                      /****
 *****************************************************************************************************/
#   define blin_internal_flags (cell->flags & BLIN_MASK)
    int32_t v = 0;

    ifBLIN_QX6("+ %u->%u", cell->inp, cell->out);
    if  (!!auwith_cadd(opd, cell)->ready) {
        v = auwith_cadd(opd, cell)->sample;
        goto out;
    }
    if  (!!rd01(opd, cell)) {
        ifBLIN_QW0("rd01");
        goto out;
    }
    if  (!auwith_cadd(opd, cell)->bchain) goto out;
    if  (auwith_cadd(opd, cell)->pobuf >= (auwith_cadd(opd, cell)->bchain->buf->size / AUWITH_MWBY)) {
/*@*/   relebchain(&auwith_cadd(opd, cell)->bchain);
        mark(cell, 2);
        auwith_cadd(opd, cell)->pobuf = 0;
        if  (!!rd01(opd, cell)) {
            ifBLIN_QW0("rd01");
            goto out;
    }   }
    if  (!!auwith_cadd(opd, cell)->bchain) {
        v = auwith_cadd(opd, cell)->bchain->buf->samples[auwith_cadd(opd, cell)->pobuf];
        if  ((!cell->f.doo) || !(opd->outach[cell->out]->flags & AUWITH_CAP_SKEW)) {
            ++auwith_cadd(opd, cell)->pobuf;
            auwith_cadd(opd, cell)->sample = v;
        } else if (!!cell->f.faser && (AUWITH_FASMIN > cell->f.faser)) {
            v = (v + auwith_cadd(opd, cell)->sample) / 2;
            ifBLIN_QX1("sample insert at %d", cell->f.faser);
        } else if (AUWITH_FASMAX < cell->f.faser) {
            auwith_cadd(opd, cell)->pobuf += 2;
            ifBLIN_QX1("sample delete at %d", cell->f.faser);
        } else {
            ++auwith_cadd(opd, cell)->pobuf;
            auwith_cadd(opd, cell)->sample = v;
        }
        cell->f.doo = 0;
        ++auwith_cadd(opd, cell)->ready;
    }
out:
    ifBLIN_QX6("- %d", v);
    return(v);
#   undef blin_internal_flags
}

static void
/*****************************************************************************************************/
take01(auwith_opd *opd, auwith_cell *cell) {                                                     /****
 *****************************************************************************************************/
#   define blin_internal_flags (cell->flags & BLIN_MASK)
    ifBLIN_QX6("+ %u->%u", cell->inp, cell->out);
    auwith_cadd(opd, cell)->ready = 0;
    ifBLIN_QX6("-");
#   undef blin_internal_flags
}

static int
/*****************************************************************************************************/
write01(auwith_opd *opd, auwith_cell *cell, auwith_buf *buf) {                                   /****
 *****************************************************************************************************/
#   define blin_internal_flags (cell->flags & BLIN_MASK)
    auwith_bchain *bchain;
    int ex = EX_OK;

    ifBLIN_QX4("+ %u->%u", cell->inp, cell->out);
    if  (!!opd->outach[cell->out]->ent->plug->rite) {
/*@*/   ex = opd->outach[cell->out]->ent->plug->rite(opd, cell, buf);
        mark(cell, 3);
    } else {
        if  (!auwith_cadd(opd, cell)->bchain) {
/*@*/       auwith_cadd(opd, cell)->bchain = takebchain(buf);
            auwith_cadd(opd, cell)->pobuf = 0;
        } else {
            for (bchain = auwith_cadd(opd, cell)->bchain; !!bchain->next; bchain = bchain->next);
/*@*/       bchain->next = takebchain(buf);
        }
        mark(cell, 1);
        if  (!!opd->outach[cell->out]->ent->plug->passiv) {
            ex = opd->outach[cell->out]->ent->plug->passiv(opd, cell, AUWITH_KIND_WRITE);
    }   }
    ifBLIN_QX4("- %d", ex);
    return(ex);
#   undef blin_internal_flags
}

/*****************************************************************************************************/
auwith_interhand                                                                                 /****
 *****************************************************************************************************/
auwith_handlers[] =
{ {NULL, NULL, NULL, NULL, NULL, NULL}
, {init01, fini01, get01, take01, read01, write01}
, {init02, fini01, get01, take01, read01, write01}
, {NULL, NULL, NULL, NULL, NULL, NULL}
};
