/*-
 * Copyright (C)2014 @BABOLO http://www.babolo.ru/
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static const char copyright[] = "\
@(#)Copyright (C)2014 @BABOLO http://www.babolo.ru/\n\
@(#)All rights reserved.\n\
@(#)V.M "VERS;
static const char rcsid[] = "$Id: plug.c,v 1.1 2014/08/14 09:34:37 babolo Exp $";
#endif /* not lint */

#define BLIN_COMPAT  3
#define Bpars_COMPAT 3

#include <sys/types.h>
#include <sys/event.h>
#include <sys/time.h>
#include <sysexits.h>
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <err.h>
#include <babolo/BLINflag.h>
#include <babolo/parser.h>
#include <babolo/recobe.h>
#include "auwith_int.h"

int
/*****************************************************************************************************
 *****************************************************************************************************/
auwith_plumb(auwith_opd *opd, auwith_io *io, int kind) {                                         /****
 *****************************************************************************************************
 *****************************************************************************************************/
    int ex = EX_OK;

#   define blin_internal_flags (opd->flags & BLIN_MASK)
    if  (!!io) {
        switch (kind) {
        case 0:
            if  (!!opd->inpach[io->sock]) {
                ifBLIN_QX0( "reconnect INP sock %u %s by %s"
                          , io->sock, opd->inpach[io->sock]->ent->plug->pname, io->ent->plug->pname
                          );
                ERROUT(EX_DATAERR, ENOTEMPTY);
            }
            opd->inpach[io->sock] = io;
            break;
        case 1:
            if  (!!opd->outach[io->sock]) {
                ifBLIN_QX0( "reconnect OUT sock %u %s by %s"
                          , io->sock, opd->outach[io->sock]->ent->plug->pname, io->ent->plug->pname
                          );
                ERROUT(EX_DATAERR, ENOTEMPTY);
            }
            opd->outach[io->sock] = io;
            break;
        default:
            ifBLIN_QX0("Ill kind %d", kind);
            ERROUT(EX_SOFTWARE, EINVAL);
    }   }
out:
    return(ex);
#   undef blin_internal_flags
}

auwith_hc *
/*****************************************************************************************************
 *****************************************************************************************************/
auwith_cadd(auwith_opd *opd, auwith_cell *cell) {                                                /****
 *****************************************************************************************************
 *****************************************************************************************************/
#   define blin_internal_flags (cell->flags & BLIN_MASK)
    auwith_hc *ref = NULL;

    if  (cell->inp >= opd->pidm->ninp) {
        ifBLIN_QX1("INP %u >= %u", cell->inp, opd->pidm->ninp);
        goto out;
    }
    if  (cell->out >= opd->pidm->nout) {
        ifBLIN_QX1("OUT %u >= %u", cell->out, opd->pidm->nout);
        goto out;
    }
    ref = &opd->ref[opd->pidm->nout * cell->inp + cell->out];
out:
    return(ref);
#   undef blin_internal_flags
}

int
/*****************************************************************************************************
 *****************************************************************************************************/
auwith_queue(auwith_opd *opd, struct kevent *ev, void *udata, auwith_time time) {                /****
 *****************************************************************************************************
 *****************************************************************************************************/
#   define blin_internal_flags (opd->flags & BLIN_MASK)
    struct timespec tn = {0, 0};
    auwith_times **t, *r;
    int ex = EX_OK;

    ifBLIN_QX3("+ %s %llu", ev ? "ev" : "", time);
    if  (!!time) {
        if  (!(r = calloc(1, sizeof(auwith_times)))) {
            ifBLIN_QX0("Malloc failed");
            ERROUT(EX_OSERR, ENOMEM);
        }
        r->time = time;
        for (t = &opd->times; !!*t && (*t)->time < time; t = &(*t)->next);
        r->udata = udata;
        r->next = *t;
        *t = r;
    }
    if  (!!ev) {
        ev->udata = udata;
        if  (!!(ex = kevent(opd->kq, ev, 1, NULL, 0, &tn))) {
            ifBLIN_QW0("kevent set");
            EXOUT(EX_IOERR);
    }   }
out:
    ifBLIN_QX3("- %d", ex);
    return(ex);
#   undef blin_internal_flags
}

int
/*****************************************************************************************************
 *****************************************************************************************************/
auwith_event(auwith_opd *opd, struct kevent *ev) {                                               /****
 *****************************************************************************************************
 *****************************************************************************************************/
#   define blin_internal_flags (opd->flags & BLIN_MASK)
    int c, ex = AUWITH_KIND_TIME;
    struct timespec tn, *tw;
    auwith_time time;
    auwith_times *t;

    if  (!(time = auwith_future(0))) {
        ifBLIN_QW0("future");
        EXOUT(-EX_OSERR);
    }
    ifBLIN_QX3("+ %llu", time);
    c = 2;
    if  (!!opd->times) {
        if  (time >= opd->times->time) c = 0; else c = 1;
    }
    if  (c & 1) {
        time = opd->times->time - time;
        tn.tv_sec = time / 1000000000ULL;
        tn.tv_nsec = time % 1000000000ULL;
        tw = &tn;
        ex = kevent(opd->kq, NULL, 0, ev, 1, tw);
        ifBLIN_QX4(" 1 %d %llu", ex, time);
    }
    if  (c & 2) {
        ex = kevent(opd->kq, NULL, 0, ev, 1, NULL);
        if  (!ex) {
            ifBLIN_QW0("kevent get %d", ex);
            ex = -EX_IOERR;
        }
        ifBLIN_QX4(" 2 %d", ex);
    } else if (!ex) {
        ev->udata = opd->times->udata;
        t = opd->times->next;
        free(opd->times);
        opd->times = t;
        ifBLIN_QX4(" 0 %d", ex);
    }
    if  (ex > 1) {
        ifBLIN_QW0("kevent get %d", ex);
        ex = -EX_IOERR;
    }
out:
    ifBLIN_QX3("- %d", ex);
    return(ex);
#   undef blin_internal_flags
}

static int64_t
/*****************************************************************************************************/
plusample(int32_t sample, auwith_cell *cell) {                                                   /****
 *****************************************************************************************************/
    int32_t level = cell->flags & AUWITH_LEVEL;
    int64_t v;

    if  (!(cell->flags & AUWITH_ALGO)) {
        v = 0;
    } else if (level == 32768) {
        v = sample;
    } else {
        v = sample * level;
        if  (v > 0) v += 16383;
        if  (v < 0) v -= 16383;
        v /= 32768;
    }
    return(v);
}

static int64_t
/*****************************************************************************************************/
bigswap(int64_t v, u_int32_t s) {                                                                 /****
 *****************************************************************************************************/
    if  (s & 0x01) v = ((v >>  1) & 0x5555555555555555ULL) | ((v <<  1) & 0xAAAAAAAAAAAAAAAAULL);
    if  (s & 0x02) v = ((v >>  2) & 0x3333333333333333ULL) | ((v <<  2) & 0xCCCCCCCCCCCCCCCCULL);
    if  (s & 0x04) v = ((v >>  4) & 0x0F0F0F0F0F0F0F0FULL) | ((v <<  4) & 0xF0F0F0F0F0F0F0F0ULL);
    if  (s & 0x08) v = ((v >>  8) & 0x00FF00FF00FF00FFULL) | ((v <<  8) & 0xFF00FF00FF00FF00ULL);
    if  (s & 0x10) v = ((v >> 16) & 0x0000FFFF0000FFFFULL) | ((v << 16) & 0xFFFF0000FFFF0000ULL);
    if  (s & 0x20) v = ((v >> 32) & 0x00000000FFFFFFFFULL) | ((v << 32) & 0xFFFFFFFF00000000ULL);
    return(v);
}

//#define HARD_PREPOINTER_DEBUG 1
auwith_buf *
/*****************************************************************************************************
 *****************************************************************************************************/
auwith_prepare(auwith_opd *opd, auwith_ent *ent) {                                               /****
 *****************************************************************************************************
 *****************************************************************************************************/
#   define blin_internal_flags (ent->flags & BLIN_MASK)
    int oblsize, i, o, multisample, samplewidth, p, pp, ex = EX_OK;
    auwith_buf *buf = NULL;
    auwith_interhand *h;
    auwith_cell *cell;
    int64_t v;
    u_int c;

    ifBLIN_QX4("+");
    samplewidth = (ent->flags & (AUWITH_CAP_ALAW | AUWITH_CAP_uLAW)) ? 1 : (ent->bits + 7) / 8;
    oblsize = samplewidth * ent->nchan * ent->block;
    if  (!(buf = auwith_creabuf(oblsize))) {                                                      /*@*/
        ifBLIN_QW0("creabuf");
        goto out;
    }
#if defined(HARD_PREPOINTER_DEBUG)
    {   auwith_bchain *bchain;
        fprintf(stderr,"p+");
        for (bchain = opd->ref[0].bchain; !!bchain; bchain = bchain->next) {
            fprintf(stderr, " %d", bchain->buf->size);
        }
        fprintf(stderr, " [%d]\n", opd->ref[0].pobuf);
    }
#endif
    for (multisample = 0, p = 0, pp = 0; multisample < ent->block; ++multisample, pp = p) {
        for (c = 0; c < ent->nchan; ++c) {
            v = 0;
            if  (!!ent->ou[c]) {
                o = ent->ou[c]->sock;
                for (i = 0; i < opd->pidm->ninp; ++i) {
                    if  (!opd->inpach[i]) continue;
                    cell = auwith_cella(opd->pidm, i, o);
                    h = &auwith_handlers[(cell->flags & AUWITH_ALGO) >> AUWITH_ALGOS];
                    if  (!h->init) continue;
                    v += plusample(h->get(opd, cell), cell);
                    if  (!auwith_cadd(opd, cell)->ready) {
                        ifBLIN_QX1( "short get %d from %"BLIN_D" [%d %d] c%u %s"
                                  , multisample, ent->block, i, o, c, ent->plug->pname
                                  );
                        goto brk;
                }   }
                if  (v > 32767) v = 32767;
                if  (v < -32767) v = -32767;
                if  (opd->outvol[o].min > v) opd->outvol[o].min = v;
                if  (opd->outvol[o].max < v) opd->outvol[o].max = v;
            }
            switch (samplewidth) {
            case 1: {
                    char *b1 = (char*)buf->samples;

                    if  (ent->flags & AUWITH_CAP_ALAW) {
                        v = recobe_G711A(v);
                    } else if (ent->flags & AUWITH_CAP_uLAW) {
                        v = recobe_G711u(v);
                    } else  {
                        v = v >> 8;
                    }
                    v = bigswap(v, ent->flags >> AUWITH_CAP_SWAPS);
                    v = v ^ ent->xormask;
                    b1[p++] = v;
                    break;
                }
            case 2: {
                    int16_t *b2 = buf->samples;

                    v = bigswap(v, ent->flags >> AUWITH_CAP_SWAPS);
                    v = v ^ ent->xormask;
                    b2[p++] = v;
                    break;
                }
            default:
                ifBLIN_QX0("cf->bits ? %u %u", ent->bits, samplewidth);
                ERROUT(EX_SOFTWARE, EINVAL);
        }   }
        for (c = 0; c < ent->nchan; ++c) {
            if  (!ent->ou[c]) continue;
            o = ent->ou[c]->sock;
            for (i = 0; i < opd->pidm->ninp; ++i) {
                if  (!opd->inpach[i]) continue;
                cell = auwith_cella(opd->pidm, i, o);
                h = &auwith_handlers[(cell->flags & AUWITH_ALGO) >> AUWITH_ALGOS];
                if  (!h->init) continue;
                h->take(opd, cell);
    }   }   }
brk:
#if defined(HARD_PREPOINTER_DEBUG)
    {   auwith_bchain *bchain;
        fprintf(stderr,"p-");
        for (bchain = opd->ref[0].bchain; !!bchain; bchain = bchain->next) {
            fprintf(stderr, " %d", bchain->buf->size);
        }
        fprintf(stderr, " [%d]\n", opd->ref[0].pobuf);
    }
#endif
    if  (buf->size > pp * samplewidth) {
/*@*/   buf->size = pp * samplewidth;
    }
out:
    ifBLIN_QX4("- %d", buf ? buf->size : -1);
    return(buf);                                                                                  /*@*/
#   undef blin_internal_flags
}

int
/*****************************************************************************************************
 *****************************************************************************************************/
auwith_postpare(auwith_opd *opd, auwith_ent *ent, void *inpb, int p) {                           /****
 *****************************************************************************************************
 *****************************************************************************************************/
#   define blin_internal_flags (ent->flags & BLIN_MASK)
    int sample, i, o, samplewidth, pp, s, ex = EX_OK;
    auwith_buf *buf[AUWITH_MACH];
    auwith_interhand *h;
    int64_t v;
    u_int c;

    ifBLIN_QX4("+ %d", p);
    for (c = 0; c < ent->nchan; ++c) {
/*@*/   if  (!(buf[c] = auwith_creabuf(ent->nchan * ent->block * AUWITH_MWBY))) {
            ifBLIN_QW0("creabuf");
            EXOUT(EX_OSERR);
        }
/*@*/   buf[c]->size = 0;
    }
    samplewidth = (ent->flags & (AUWITH_CAP_ALAW | AUWITH_CAP_uLAW)) ? 1 : (ent->bits + 7) / 8;
    for (sample = 0, s = 0, pp = 0; sample < ent->block; ++sample) {
        if  (s >= p) break;
        for (c = 0; c < ent->nchan; ++c) {
        ;   if  (s >= p) break;
        ;   switch (samplewidth) {
            case 1: {
                    char *b1 = inpb;

        ;           v = b1[pp++];
        ;           v = v ^ ent->xormask;
        ;           v = bigswap(v, ent->flags >> AUWITH_CAP_SWAPS);
        ;           if  (ent->flags & AUWITH_CAP_ALAW) {
        ;               v = recobe_deG711A[v];
        ;           } else if (ent->flags & AUWITH_CAP_uLAW) {
        ;               v = recobe_deG711u[v];
        ;           } else  {
        ;               v = v << 8;
        ;           }
        ;           ++s;
        ;           break;
        ;       }
            case 2: {
                    int16_t *b2 = inpb;

        ;           v = b2[pp++];
        ;           v = v ^ ent->xormask;
        ;           v = bigswap(v, ent->flags >> AUWITH_CAP_SWAPS);
        ;           s += 2;
        ;           break;
        ;       }
            default:
        ;       ifBLIN_QX0("cf->bits ? %u %u", ent->bits, samplewidth);
        ;       ERROUT(EX_SOFTWARE, EINVAL);
        ;   }
/*@*/   ;   buf[c]->samples[buf[c]->size++] = v;
        ;   if  (!!ent->in[c]) {
        ;       i = ent->in[c]->sock;
        ;       if  (opd->inpvol[i].min > v) opd->inpvol[i].min = v;
        ;       if  (opd->inpvol[i].max < v) opd->inpvol[i].max = v;
        }   }
        if  (s >= p) break;
    }
    for (c = 0; c < ent->nchan; ++c) {
/*@*/   buf[c]->size *= 2;
        if  (!ent->in[c]) continue;
        i = ent->in[c]->sock;
        for (o = 0; o < opd->pidm->nout; ++o) {
            if  (!opd->outach[o]) continue;
            h = &auwith_handlers[(auwith_cella(opd->pidm, i, o)->flags & AUWITH_ALGO) >> AUWITH_ALGOS];
            if  (  !!(auwith_cella(opd->pidm, i, o)->flags & AUWITH_ALGO)
/*@*/           && !!(ex = h->write(opd, auwith_cella(opd->pidm, i, o), buf[c]))
                ) {
                ifBLIN_QW0("h->write %d %d", i, o);
                EXOUT(EX_IOERR);
        }   }
/*@*/   auwith_relebuf(buf[c]);
/*@*/   buf[c] = NULL;
    }
out:
    ifBLIN_QX4("- %d", ex);
    return(ex);
#   undef blin_internal_flags
}

