/*-
 * Copyright (C)2014 @BABOLO http://www.babolo.ru/
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static const char copyright[] = "\
@(#)Copyright (C)2014 @BABOLO http://www.babolo.ru/\n\
@(#)All rights reserved.\n\
@(#)V.M "VERS;
static const char rcsid[] = "$Id: udpin.c,v 1.1 2014/08/14 09:34:37 babolo Exp $";
#endif /* not lint */

#define BLIN_COMPAT  3
#define Bpars_COMPAT 3

#include <sys/types.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <sys/event.h>
#include <sys/time.h>
#include <sysexits.h>
#include <strings.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <netdb.h>
#include <errno.h>
#include <err.h>
#include <babolo/BLINflag.h>
#include <babolo/parser.h>
#include <babolo/recobe.h>
#include "auwith_int.h"
                        /*****************************/
typedef struct {        /********* auwith_ent_udpi ***/
    auwith_ent      common;
    struct addrinfo*ai;                         /* addrinfo                            */
    char           *tport;                      /*                       */
    int             sock;                       /* UDP                            */
    void           *inpb;
} auwith_ent_udpi;
                        /*****************************/
typedef struct {        /********* auwith_io_udpi ****/
    auwith_io       common;
    u_int16_t       chan;
} auwith_io_udpi;

static auwith_time sta = 0;

static void
/*****************************************************************************************************/
usage(void) {                                                                                    /****
 *****************************************************************************************************/
    fprintf( stderr
           , "auwithd udpi V%u @BABOLO " DATE "\n"
             "Usage: auwithd [-options] udpi [-options]\n"
             "   -A              A law\n"
             "   -b num          sample width in bits (default %u)\n"
             "   -c num          channels number\n"
             "   -D              hard timing debug\n"
             "   -E -X XX        use XX as env name for opt's arg\n"
             "   -h [host]:port  listen for UDP stream\n"
             "   -i in           matrix's input number\n"
             "   -L              local phase count\n"
             "   -q              silent\n"
             "   -r freq         samples rate in Hz (default %u)\n"
             "   -s num          block size in samples (default %u)\n"
             "   -u              u law\n"
             "   -v[v[v...]]     verbose\n"
             "   -w 0xXX         swap bits (0x01), doubles (0x02), nibbles (0x04) so on\n"
             "   -z TEXT         parce TEXT as arguments continuation\n"
             "   -Z TEXT         parce TEXT as arguments continuation with env substitution\n"
             "   -= mode         internal diagnostic\n"
             "   -?              this text\n"
           , auwith_def_udpi.mark.ver
           , AUWITH_BITS
           , AUWITH_RATE
           , AUWITH_BLOCK
           )
    ;
}

static int
/*****************************************************************************************************/
parsin(auwith_opd *opd) {                                                                        /****
 *****************************************************************************************************/
    struct kevent ev = {0, EVFILT_READ, EV_ADD, 0, 0, NULL};
    char *flar = "Ab:c:Dh:i:Lqr:s:uvw:=:?";
    struct addrinfo *hint;
    auwith_ent_udpi *ent;
    int c, ex = EX_OK;
    u_int16_t inpp;
    char *q;

    ifBLIN_QX3("+");
    if  (!opd) {
        ifBLIN_QX0("no opd");
        ERROUT(EX_DATAERR, EFAULT);
    }
#   define blin_internal_flags (opd->flags & BLIN_MASK)
    if  (!opd->bos) {
        ifBLIN_QX0("no bos");
        ERROUT(EX_DATAERR, EFAULT);
    }
    if  (!(ent = calloc(1, sizeof(auwith_ent_udpi)))) {
        ifBLIN_QX0("Malloc failed #1");
        ERROUT(EX_OSERR, ENOMEM);
    }
    ent->common.flags = (opd->flags & ~AUWITH_CAP_MASK) | (auwith_def_dspo.flags & AUWITH_CAP_MASK);
    ent->common.plug = &auwith_def_udpi;
    ent->common.block = AUWITH_BLOCK;
    ent->common.bits = AUWITH_BITS;
    ent->common.rate = AUWITH_RATE;
    if  (!!babolo_chopts(opd->bos, 0, flar)) {
        ifBLIN_QW0("chopts");
        EXOUT(EX_SOFTWARE);
    }
    ex = EX_OK;
    if  (opd->flags & AUWITH_DUMPOP) babolo_dumpopts(opd->bos);
    while (!!(c = babolo_getopts(opd->bos))) {
        switch (c) {
        case 'A': ent->common.flags |= AUWITH_CAP_ALAW;
                ; break;
        case 'b': ent->common.bits = babolo_getoptunum(opd->bos, 0);
                ; break;
        case 'c': if  (!!ent->common.nchan) {
                      ifBLIN_QX0("redefine nchan");
                      ERROUT(EX_USAGE, ENOTEMPTY);
                  }
                ; ent->common.nchan = babolo_getoptunum(opd->bos, 0);
                ; inpp = 0;
                ; break;
        case 'D': ent->common.flags |= AUWITH_HARDTI;
                ; break;
        case 'h': if  (!(q = strdup(babolo_getoptsarg(opd->bos)))) {
                      ifBLIN_QW0("no -h arg");
                      ERROUT(EX_USAGE, EINVAL);
                  }
                ; if  (!(ent->tport = rindex(q, ':'))) {
                      ifBLIN_QW0("no -h :port");
                      ERROUT(EX_USAGE, EINVAL);
                  }
                ; *(ent->tport++) = '\0';
                ; if  (!(hint = calloc(1, sizeof(*hint)))) {
                      ifBLIN_QX0("Malloc failed #2");
                      ERROUT(EX_OSERR, ENOMEM);
                  }
                ; hint->ai_family = PF_INET;
                ; hint->ai_socktype = SOCK_DGRAM;
                ; hint->ai_protocol = IPPROTO_UDP;
                ; hint->ai_flags = AI_NUMERICSERV | AI_ADDRCONFIG | AI_CANONNAME | AI_PASSIVE;
                ; if  (!*q) q = NULL;
                ; if  ((ex = getaddrinfo(q, ent->tport, hint, &ent->ai))) {
                      ifBLIN_QX0("getaddrinfo: %s: %s", opd->bos->o, gai_strerror(ex));
                      ERROUT(EX_OSERR, ENETUNREACH);
                  }
                ; free(hint);
                ; hint = NULL;
                ; ent->sock = socket(ent->ai->ai_family, ent->ai->ai_socktype, ent->ai->ai_protocol);
                ; if  (0 > ent->sock) {
                      ifBLIN_QW0("socket");
                      EXOUT(EX_IOERR);
                  }
                ; if  (0 > bind(ent->sock, ent->ai->ai_addr, ent->ai->ai_addrlen)) {
                      ifBLIN_QW0("bind");
                      EXOUT(EX_IOERR);
                  }
                ; break;
        case 'i': if  (inpp >= ent->common.nchan) {
                      ifBLIN_QX0("chan %u >= %u", inpp, ent->common.nchan);
                      EXOUT(EX_USAGE);
                  }
                ; if  (!(ent->common.in[inpp] = calloc(1, sizeof(auwith_io_udpi)))) {
                      ifBLIN_QX0("Malloc failed #3");
                      ERROUT(EX_OSERR, ENOMEM);
                  }
                ; ent->common.in[inpp]->flags = ent->common.flags;
                ; ent->common.in[inpp]->ent = (auwith_ent*)ent;
                ; ((auwith_io_udpi*)ent->common.in[inpp])->chan = inpp;
                ; ent->common.in[inpp]->sock = babolo_getoptunum(opd->bos, 0);
                ; if  (ent->common.in[inpp]->sock >= opd->inpachsz) {
                      ifBLIN_QX0( "-i[%u] must be < %u, now %u"
                                , inpp, opd->inpachsz, ent->common.in[inpp]->sock
                                );
                      ERROUT(EX_USAGE, EINVAL);
                  }
                ; ++inpp;
                ; break;
        case 'L': ent->common.flags ^= AUWITH_CFAF;
                ; break;
        case 'q': BLIN_QUIET(ent->common.flags);
                ; break;
        case 'r': ent->common.rate = babolo_getoptunum(opd->bos, 0);
                ; break;
        case 's': ent->common.block = babolo_getoptunum(opd->bos, 0);
                ; break;
        case 'u': ent->common.flags |= AUWITH_CAP_uLAW;
                ; break;
        case 'v': BLIN_VERBOSE(ent->common.flags);
                ; break;
        case 'w': ent->common.flags &= ~AUWITH_CAP_SWAP;
                ; ent->common.flags |= (babolo_getoptunum(opd->bos, 0) << 6) & AUWITH_CAP_SWAP;
                ; break;
        case '=': switch(babolo_getoptsarg(opd->bos)[0]) {
                  case '0': blin_ctl(BLIN_CTL_DUMP);           exit(EX_OK);
                  case '1': babolo_optest(flar);               exit(EX_OK);
                  case '2': babolo_optext(opd->bos);           exit(EX_OK);
                  case '3': babolo_dumpopts(opd->bos);         exit(EX_OK);
                  }
                ; break;
        case '?': usage();
                ; break;
        default : ifBLIN_QX0("Illegal flag -%c", opd->bos->c);
                ; usage();
                ; ERROUT(EX_USAGE, EINVAL);
    }   }
    if  (!ent->common.bits || (ent->common.bits > 16)) {
        ifBLIN_QX0("-b must be >0 and <=16, now %u", ent->common.bits);
        ERROUT(EX_USAGE, EINVAL);
    }
    if  (!ent->common.rate || (ent->common.rate > 48000)) {
        ifBLIN_QX0("-r must be >0 and <=48000, now %u", ent->common.rate);
        ERROUT(EX_USAGE, EINVAL);
    }
    for (inpp = 0; !ex && (inpp < ent->common.nchan); ++inpp) {
        if  (!!(ex = auwith_plumb(opd, (auwith_io*)ent->common.in[inpp], 0))) {
            ifBLIN_QX0("auwith_plumbi");
            goto out;
    }   }
    if  (!(ent->inpb = malloc(AUWITH_PKTSZ))) {
        ifBLIN_QX0("Malloc failed #4");
        ERROUT(EX_OSERR, ENOMEM);
    }
    ev.ident = ent->sock;
    if  (!!(ex = auwith_queue(opd, &ev, &ent->common, 0))) {
        ifBLIN_QW0("queue");
        goto out;
    }
out:
    if  (!!ex && !!ent) {
        for (inpp = 0; inpp < ent->common.nchan; ++inpp) {
            if  (!!ent->common.in[inpp]) free(ent->common.in[inpp]);
        }
        free(ent);
    }
    ifBLIN_QX3("- %d", ex);
    return(ex);
#   undef blin_internal_flags
}

static void
/*****************************************************************************************************/
deconf(auwith_opd *opd, auwith_io *ioc, FILE *fo) {                                              /****
 *****************************************************************************************************/
    auwith_io_udpi *io = (auwith_io_udpi*)ioc;

#   define blin_internal_flags (ioc->flags & BLIN_MASK)
    fprintf( fo, "[%u] udpi %uHz, %usamples, chan %u, %s:%s(%d)%s, sw=%02X^%08X\n"
           , io->common.sock
           , io->common.ent->rate
           , io->common.ent->block
           , io->chan
           , ((auwith_ent_udpi*)io->common.ent)->ai->ai_canonname
           , ((auwith_ent_udpi*)io->common.ent)->tport
           , ((auwith_ent_udpi*)io->common.ent)->sock
           ,   (io->common.ent->flags & AUWITH_CAP_ALAW) ? ", G711A"
             : (io->common.ent->flags & AUWITH_CAP_uLAW) ? ", G711u" : ""
           , (io->common.ent->flags & AUWITH_CAP_SWAP) >> AUWITH_CAP_SFULLS
           , io->common.ent->xormask
           );
    fprintf(fo, "    ");
    auwith_capand(opd->flags, fo, 0);
    fprintf(fo, " ");
    auwith_capand(ioc->ent->flags, fo, 1);
    fprintf(fo, " ");
    auwith_capand(ioc->flags, fo, 1);
    fprintf(fo, "\n");
#   undef blin_internal_flags
}

static int
/*****************************************************************************************************/
activ(auwith_opd *opd, auwith_ent *ent, int kind) {                                              /****
 *****************************************************************************************************/
    auwith_ent_udpi *cf = (auwith_ent_udpi*)ent;
    int iblsize, samplewidth, p, ex = EX_OK;
    auwith_time tt;

#   define blin_internal_flags (ent->flags & BLIN_MASK)
    ifBLIN_QX3("+ %d", kind);
    if  (kind != AUWITH_KIND_KQ) {
        ifBLIN_QX0("kind = %d", kind);
        ERROUT(EX_SOFTWARE, EINVAL);
    }
    samplewidth = (ent->flags & (AUWITH_CAP_ALAW | AUWITH_CAP_uLAW)) ? 1 : (ent->bits + 7) / 8;
    iblsize = samplewidth * ent->nchan * ent->block;
    auwith_Tmark(opd, 10);
    p = read(cf->sock, cf->inpb, AUWITH_PKTSZ);
    auwith_Tmark(opd, 11);
    if  (0 > p) {
        ifBLIN_QW0("read");
        EXOUT(EX_IOERR);
    } else if (p > AUWITH_PKTSZ) {
        ifBLIN_QX0("read %d > %d", p, AUWITH_PKTSZ);
        ERROUT(EX_IOERR, EIO);
    } else if (p != iblsize) {
        ifBLIN_QX1("read %d <> %d", p, iblsize);
    }
    if  (!!(ex = auwith_postpare(opd, ent, cf->inpb, p))) {
        ifBLIN_QW0("auwith_postpare");
        goto out;
    }
    if  (ent->flags & AUWITH_HARDTI) {
        tt = auwith_future(0);
        fprintf(stderr, "%u . %d %llu\n", iblsize, p, (tt - sta) / 1000);
        sta = tt;
    }
out:
    ifBLIN_QX3("- %d", ex);
    return(ex);
#   undef blin_internal_flags
}

/*****************************************************************************************************
 *****************************************************************************************************/
auwith_plug auwith_def_udpi =                                                                    /****
 *****************************************************************************************************
 *****************************************************************************************************/
{ "udpi"
, AUWITH_MARK
, parsin
, deconf
, activ
, NULL
, NULL
, NULL
, AUWITH_CAP_INP
, 0
};
